import json
from datetime import timedelta

from django.utils import timezone

from .. import test
from ...categories.models import Category
from ...readtracker import poststracker
from ..models import Thread
from ..test import patch_category_acl, patch_other_category_acl
from .test_threads_api import ThreadsApiTestCase


class ThreadPatchApiTestCase(ThreadsApiTestCase):
    def patch(self, api_link, ops):
        return self.client.patch(
            api_link, json.dumps(ops), content_type="application/json"
        )


class ThreadAddAclApiTests(ThreadPatchApiTestCase):
    def test_add_acl_true(self):
        """api adds current thread's acl to response"""
        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertTrue(response_json["acl"])

    def test_add_acl_false(self):
        """
        if value is false, api won't add acl to the response, but will set empty key
        """
        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIsNone(response_json["acl"])


class ThreadChangeTitleApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_edit_threads": 2})
    def test_change_thread_title(self):
        """api makes it possible to change thread title"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "title", "value": "Lorem ipsum change!"}],
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["title"], "Lorem ipsum change!")

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["title"], "Lorem ipsum change!")

    @patch_category_acl({"can_edit_threads": 0})
    def test_change_thread_title_no_permission(self):
        """api validates permission to change title"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "title", "value": "Lorem ipsum change!"}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't edit threads in this category."
        )

    @patch_category_acl({"can_edit_threads": 2, "can_close_threads": 0})
    def test_change_thread_title_closed_category_no_permission(self):
        """api test permission to edit thread title in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "title", "value": "Lorem ipsum change!"}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't edit threads in it.",
        )

    @patch_category_acl({"can_edit_threads": 2, "can_close_threads": 0})
    def test_change_thread_title_closed_thread_no_permission(self):
        """api test permission to edit closed thread title"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "title", "value": "Lorem ipsum change!"}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This thread is closed. You can't edit it."
        )

    @patch_category_acl({"can_edit_threads": 1, "thread_edit_time": 1})
    def test_change_thread_title_after_edit_time(self):
        """api cleans, validates and rejects too short title"""
        self.thread.started_on = timezone.now() - timedelta(minutes=10)
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "title", "value": "Lorem ipsum change!"}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't edit threads that are older than 1 minute.",
        )

    @patch_category_acl({"can_edit_threads": 2})
    def test_change_thread_title_invalid(self):
        """api cleans, validates and rejects too short title"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "title", "value": 12}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "Thread title should be at least 5 characters long (it has 2).",
        )


class ThreadPinGloballyApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_pin_threads": 2})
    def test_pin_thread(self):
        """api makes it possible to pin globally thread"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 2}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["weight"], 2)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 2)

    @patch_category_acl({"can_pin_threads": 2, "can_close_threads": 0})
    def test_pin_thread_closed_category_no_permission(self):
        """api checks if category is closed"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 2}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't change threads weights in it.",
        )

    @patch_category_acl({"can_pin_threads": 2, "can_close_threads": 0})
    def test_pin_thread_closed_no_permission(self):
        """api checks if thread is closed"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 2}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This thread is closed. You can't change its weight.",
        )

    @patch_category_acl({"can_pin_threads": 2})
    def test_unpin_thread(self):
        """api makes it possible to unpin thread"""
        self.thread.weight = 2
        self.thread.save()

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 2)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 0}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["weight"], 0)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 0)

    @patch_category_acl({"can_pin_threads": 1})
    def test_pin_thread_no_permission(self):
        """api pin thread globally with no permission fails"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 2}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't pin threads globally in this category.",
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 0)

    @patch_category_acl({"can_pin_threads": 1})
    def test_unpin_thread_no_permission(self):
        """api unpin thread with no permission fails"""
        self.thread.weight = 2
        self.thread.save()

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 2)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 1}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't change globally pinned threads weights in this category.",
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 2)


class ThreadPinLocallyApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_pin_threads": 1})
    def test_pin_thread(self):
        """api makes it possible to pin locally thread"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 1}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["weight"], 1)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 1)

    @patch_category_acl({"can_pin_threads": 1})
    def test_unpin_thread(self):
        """api makes it possible to unpin thread"""
        self.thread.weight = 1
        self.thread.save()

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 1)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 0}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["weight"], 0)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 0)

    @patch_category_acl({"can_pin_threads": 0})
    def test_pin_thread_no_permission(self):
        """api pin thread locally with no permission fails"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 1}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't change threads weights in this category.",
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 0)

    @patch_category_acl({"can_pin_threads": 0})
    def test_unpin_thread_no_permission(self):
        """api unpin thread with no permission fails"""
        self.thread.weight = 1
        self.thread.save()

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 1)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "weight", "value": 0}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't change threads weights in this category.",
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["weight"], 1)


class ThreadMoveApiTests(ThreadPatchApiTestCase):
    def setUp(self):
        super().setUp()

        Category(name="Other category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.dst_category = Category.objects.get(slug="other-category")

    @patch_other_category_acl({"can_start_threads": 2})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_no_top(self):
        """api moves thread to other category, sets no top category"""
        response = self.patch(
            self.api_link,
            [
                {"op": "replace", "path": "category", "value": self.dst_category.pk},
                {"op": "add", "path": "top-category", "value": self.dst_category.pk},
                {"op": "replace", "path": "flatten-categories", "value": None},
            ],
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertEqual(reponse_json["category"], self.dst_category.pk)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.dst_category.pk)

    @patch_other_category_acl({"can_start_threads": 2})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_with_top(self):
        """api moves thread to other category, sets top"""
        response = self.patch(
            self.api_link,
            [
                {"op": "replace", "path": "category", "value": self.dst_category.pk},
                {
                    "op": "add",
                    "path": "top-category",
                    "value": Category.objects.root_category().pk,
                },
                {"op": "replace", "path": "flatten-categories", "value": None},
            ],
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertEqual(reponse_json["category"], self.dst_category.pk)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.dst_category.pk)

    @patch_other_category_acl({"can_start_threads": 2})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_reads(self):
        """api moves thread reads together with thread"""
        poststracker.save_read(self.user, self.thread.first_post)

        self.assertEqual(self.user.postread_set.count(), 1)
        self.user.postread_set.get(category=self.category)

        response = self.patch(
            self.api_link,
            [
                {"op": "replace", "path": "category", "value": self.dst_category.pk},
                {"op": "add", "path": "top-category", "value": self.dst_category.pk},
                {"op": "replace", "path": "flatten-categories", "value": None},
            ],
        )
        self.assertEqual(response.status_code, 200)

        # thread read was moved to new category
        postreads = self.user.postread_set.filter(post__is_event=False).order_by("id")

        self.assertEqual(postreads.count(), 1)
        postreads.get(category=self.dst_category)

    @patch_other_category_acl({"can_start_threads": 2})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_subscriptions(self):
        """api moves thread subscriptions together with thread"""
        self.user.subscription_set.create(
            thread=self.thread,
            category=self.thread.category,
            last_read_on=self.thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(category=self.category)

        response = self.patch(
            self.api_link,
            [
                {"op": "replace", "path": "category", "value": self.dst_category.pk},
                {"op": "add", "path": "top-category", "value": self.dst_category.pk},
                {"op": "replace", "path": "flatten-categories", "value": None},
            ],
        )
        self.assertEqual(response.status_code, 200)

        # thread read was moved to new category
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(category=self.dst_category)

    @patch_category_acl({"can_move_threads": False})
    def test_move_thread_no_permission(self):
        """api move thread to other category with no permission fails"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.dst_category.pk}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't move threads in this category."
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.category.pk)

    @patch_other_category_acl({"can_close_threads": False})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_closed_category_no_permission(self):
        """api move thread from closed category with no permission fails"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.dst_category.pk}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't move it's threads.",
        )

    @patch_other_category_acl({"can_close_threads": False})
    @patch_category_acl({"can_move_threads": True})
    def test_move_closed_thread_no_permission(self):
        """api move closed thread with no permission fails"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.dst_category.pk}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This thread is closed. You can't move it."
        )

    @patch_other_category_acl({"can_see": False})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_no_category_access(self):
        """api move thread to category with no access fails"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.dst_category.pk}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(response_json["detail"][0], "NOT FOUND")

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.category.pk)

    @patch_other_category_acl({"can_browse": False})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_no_category_browse(self):
        """api move thread to category with no browsing access fails"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.dst_category.pk}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            'You don\'t have permission to browse "Other category" contents.',
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.category.pk)

    @patch_other_category_acl({"can_start_threads": False})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_no_category_start_threads(self):
        """api move thread to category with no posting access fails"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.dst_category.pk}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You don't have permission to start new threads in this category.",
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.category.pk)

    @patch_other_category_acl({"can_start_threads": 2})
    @patch_category_acl({"can_move_threads": True})
    def test_move_thread_same_category(self):
        """api move thread to category it's already in fails"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "category", "value": self.thread.category_id}],
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't move thread to the category it's already in.",
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["category"]["id"], self.category.pk)

    def test_thread_flatten_categories(self):
        """api flatten thread categories"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "flatten-categories", "value": None}],
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["category"], self.category.pk)


class ThreadCloseApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_close_threads": True})
    def test_close_thread(self):
        """api makes it possible to close thread"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-closed", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertTrue(response_json["is_closed"])

        thread_json = self.get_thread_json()
        self.assertTrue(thread_json["is_closed"])

    @patch_category_acl({"can_close_threads": True})
    def test_open_thread(self):
        """api makes it possible to open thread"""
        self.thread.is_closed = True
        self.thread.save()

        thread_json = self.get_thread_json()
        self.assertTrue(thread_json["is_closed"])

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-closed", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertFalse(response_json["is_closed"])

        thread_json = self.get_thread_json()
        self.assertFalse(thread_json["is_closed"])

    @patch_category_acl({"can_close_threads": False})
    def test_close_thread_no_permission(self):
        """api close thread with no permission fails"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-closed", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You don't have permission to close this thread.",
        )

        thread_json = self.get_thread_json()
        self.assertFalse(thread_json["is_closed"])

    @patch_category_acl({"can_close_threads": False})
    def test_open_thread_no_permission(self):
        """api open thread with no permission fails"""
        self.thread.is_closed = True
        self.thread.save()

        thread_json = self.get_thread_json()
        self.assertTrue(thread_json["is_closed"])

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-closed", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You don't have permission to open this thread."
        )

        thread_json = self.get_thread_json()
        self.assertTrue(thread_json["is_closed"])


class ThreadApproveApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_approve_content": True})
    def test_approve_thread(self):
        """api makes it possible to approve thread"""
        self.thread.first_post.is_unapproved = True
        self.thread.first_post.save()

        self.thread.synchronize()
        self.thread.save()

        self.assertTrue(self.thread.is_unapproved)
        self.assertTrue(self.thread.has_unapproved_posts)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertFalse(response_json["is_unapproved"])
        self.assertFalse(response_json["has_unapproved_posts"])

        thread_json = self.get_thread_json()
        self.assertFalse(thread_json["is_unapproved"])
        self.assertFalse(thread_json["has_unapproved_posts"])

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

    @patch_category_acl({"can_approve_content": True, "can_close_threads": False})
    def test_approve_thread_category_closed_no_permission(self):
        """api checks permission for approving threads in closed categories"""
        self.thread.first_post.is_unapproved = True
        self.thread.first_post.save()

        self.thread.synchronize()
        self.thread.save()

        self.assertTrue(self.thread.is_unapproved)
        self.assertTrue(self.thread.has_unapproved_posts)

        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't approve threads in it.",
        )

    @patch_category_acl({"can_approve_content": True, "can_close_threads": False})
    def test_approve_thread_closed_no_permission(self):
        """api checks permission for approving posts in closed categories"""
        self.thread.first_post.is_unapproved = True
        self.thread.first_post.save()

        self.thread.synchronize()
        self.thread.save()

        self.assertTrue(self.thread.is_unapproved)
        self.assertTrue(self.thread.has_unapproved_posts)

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This thread is closed. You can't approve it."
        )

    @patch_category_acl({"can_approve_content": True})
    def test_unapprove_thread(self):
        """api returns permission error on approval removal"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "Content approval can't be reversed."
        )


class ThreadHideApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_hide_threads": True})
    def test_hide_thread(self):
        """api makes it possible to hide thread"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertTrue(reponse_json["is_hidden"])

        thread_json = self.get_thread_json()
        self.assertTrue(thread_json["is_hidden"])

    @patch_category_acl({"can_hide_threads": False})
    def test_hide_thread_no_permission(self):
        """api hide thread with no permission fails"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't hide threads in this category."
        )

        thread_json = self.get_thread_json()
        self.assertFalse(thread_json["is_hidden"])

    @patch_category_acl({"can_hide_threads": False, "can_hide_own_threads": True})
    def test_hide_non_owned_thread(self):
        """api forbids non-moderator from hiding other users threads"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't hide other users theads in this category.",
        )

    @patch_category_acl(
        {"can_hide_threads": False, "can_hide_own_threads": True, "thread_edit_time": 1}
    )
    def test_hide_owned_thread_no_time(self):
        """api forbids non-moderator from hiding other users threads"""
        self.thread.started_on = timezone.now() - timedelta(minutes=5)
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't hide threads that are older than 1 minute.",
        )

    @patch_category_acl({"can_hide_threads": True, "can_close_threads": False})
    def test_hide_closed_category_no_permission(self):
        """api test permission to hide thread in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't hide threads in it.",
        )

    @patch_category_acl({"can_hide_threads": True, "can_close_threads": False})
    def test_hide_closed_thread_no_permission(self):
        """api test permission to hide closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This thread is closed. You can't hide it."
        )


class ThreadUnhideApiTests(ThreadPatchApiTestCase):
    def setUp(self):
        super().setUp()

        self.thread.is_hidden = True
        self.thread.save()

    @patch_category_acl({"can_hide_threads": True})
    def test_unhide_thread(self):
        """api makes it possible to unhide thread"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertFalse(reponse_json["is_hidden"])

        thread_json = self.get_thread_json()
        self.assertFalse(thread_json["is_hidden"])

    @patch_category_acl({"can_hide_threads": False})
    def test_unhide_thread_no_permission(self):
        """api unhide thread with no permission fails as thread is invisible"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_hide_threads": True, "can_close_threads": False})
    def test_unhide_closed_category_no_permission(self):
        """api test permission to unhide thread in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't reveal threads in it.",
        )

    @patch_category_acl({"can_hide_threads": True, "can_close_threads": False})
    def test_unhide_closed_thread_no_permission(self):
        """api test permission to unhide closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This thread is closed. You can't reveal it."
        )


class ThreadMarkBestAnswerApiTests(ThreadPatchApiTestCase):
    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer(self):
        """api makes it possible to mark best answer"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["ok"],
                "best_answer": best_answer.id,
                "best_answer_is_protected": False,
                "best_answer_marked_on": response.json()["best_answer_marked_on"],
                "best_answer_marked_by": self.user.id,
                "best_answer_marked_by_name": self.user.username,
                "best_answer_marked_by_slug": self.user.slug,
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], best_answer.id)
        self.assertEqual(thread_json["best_answer_is_protected"], False)
        self.assertEqual(
            thread_json["best_answer_marked_on"],
            response.json()["best_answer_marked_on"],
        )
        self.assertEqual(thread_json["best_answer_marked_by"], self.user.id)
        self.assertEqual(thread_json["best_answer_marked_by_name"], self.user.username)
        self.assertEqual(thread_json["best_answer_marked_by_slug"], self.user.slug)

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_anonymous(self):
        """api validates that user is authenticated before marking best answer"""
        self.logout_user()

        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 0})
    def test_mark_best_answer_no_permission(self):
        """api validates permission to mark best answers"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to mark best answers "
                    'in the "First category" category.'
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 1})
    def test_mark_best_answer_not_thread_starter(self):
        """api validates permission to mark best answers in owned thread"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to mark best answer in this thread "
                    "because you didn't start it."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

        # passing scenario is possible
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_mark_best_answers": 2, "can_close_threads": False})
    def test_mark_best_answer_category_closed_no_permission(self):
        """api validates permission to mark best answers in closed category"""
        best_answer = test.reply_thread(self.thread)

        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to mark best answer in this thread "
                    'because its category "First category" is closed.'
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2, "can_close_threads": True})
    def test_mark_best_answer_category_closed(self):
        """api validates permission to mark best answers in closed category"""
        best_answer = test.reply_thread(self.thread)

        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_mark_best_answers": 2, "can_close_threads": False})
    def test_mark_best_answer_thread_closed_no_permission(self):
        """api validates permission to mark best answers in closed thread"""
        best_answer = test.reply_thread(self.thread)

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You can't mark best answer in this thread because it's closed and "
                    "you don't have permission to open it."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2, "can_close_threads": True})
    def test_mark_best_answer_thread_closed(self):
        """api validates permission to mark best answers in closed thread"""
        best_answer = test.reply_thread(self.thread)

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_invalid_post_id(self):
        """api validates that post id is int"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": "d7sd89a7d98sa"}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.id, "detail": ["A valid integer is required."]},
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_post_not_found(self):
        """api validates that post exists"""
        response = self.patch(
            self.api_link,
            [
                {
                    "op": "replace",
                    "path": "best-answer",
                    "value": self.thread.last_post_id + 1,
                }
            ],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"id": self.thread.id, "detail": ["NOT FOUND"]}
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_post_invisible(self):
        """api validates post visibility to action author"""
        unapproved_post = test.reply_thread(self.thread, is_unapproved=True)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": unapproved_post.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"id": self.thread.id, "detail": ["NOT FOUND"]}
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_post_other_thread(self):
        """api validates post belongs to same thread"""
        other_thread = test.post_thread(self.category)

        response = self.patch(
            self.api_link,
            [
                {
                    "op": "replace",
                    "path": "best-answer",
                    "value": other_thread.first_post_id,
                }
            ],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"id": self.thread.id, "detail": ["NOT FOUND"]}
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_event_id(self):
        """api validates that post is not an event"""
        best_answer = test.reply_thread(self.thread)
        best_answer.is_event = True
        best_answer.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["Events can't be marked as best answers."],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_first_post(self):
        """api validates that post is not a first post in thread"""
        response = self.patch(
            self.api_link,
            [
                {
                    "op": "replace",
                    "path": "best-answer",
                    "value": self.thread.first_post_id,
                }
            ],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["First post in a thread can't be marked as best answer."],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_hidden_post(self):
        """api validates that post is not hidden"""
        best_answer = test.reply_thread(self.thread, is_hidden=True)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["Hidden posts can't be marked as best answers."],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2})
    def test_mark_best_answer_unapproved_post(self):
        """api validates that post is not unapproved"""
        best_answer = test.reply_thread(
            self.thread, poster=self.user, is_unapproved=True
        )

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["Unapproved posts can't be marked as best answers."],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2, "can_protect_posts": False})
    def test_mark_best_answer_protected_post_no_permission(self):
        """api respects post protection"""
        best_answer = test.reply_thread(self.thread, is_protected=True)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to mark this post as best answer "
                    "because a moderator has protected it."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])

    @patch_category_acl({"can_mark_best_answers": 2, "can_protect_posts": True})
    def test_mark_best_answer_protected_post(self):
        """api respects post protection"""
        best_answer = test.reply_thread(self.thread, is_protected=True)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)


class ThreadChangeBestAnswerApiTests(ThreadPatchApiTestCase):
    def setUp(self):
        super().setUp()

        self.best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, self.best_answer)
        self.thread.save()

    @patch_category_acl({"can_mark_best_answers": 2, "can_change_marked_answers": 2})
    def test_change_best_answer(self):
        """api makes it possible to change best answer"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["ok"],
                "best_answer": best_answer.id,
                "best_answer_is_protected": False,
                "best_answer_marked_on": response.json()["best_answer_marked_on"],
                "best_answer_marked_by": self.user.id,
                "best_answer_marked_by_name": self.user.username,
                "best_answer_marked_by_slug": self.user.slug,
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], best_answer.id)
        self.assertEqual(thread_json["best_answer_is_protected"], False)
        self.assertEqual(
            thread_json["best_answer_marked_on"],
            response.json()["best_answer_marked_on"],
        )
        self.assertEqual(thread_json["best_answer_marked_by"], self.user.id)
        self.assertEqual(thread_json["best_answer_marked_by_name"], self.user.username)
        self.assertEqual(thread_json["best_answer_marked_by_slug"], self.user.slug)

    @patch_category_acl({"can_mark_best_answers": 2, "can_change_marked_answers": 2})
    def test_change_best_answer_same_post(self):
        """api validates if new best answer is same as current one"""
        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["This post is already marked as thread's best answer."],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 2})
    def test_change_best_answer_no_permission_to_mark(self):
        """
        api validates permission to mark best answers before allowing answer change
        """
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to mark best answers in the "
                    '"First category" category.'
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 2, "can_change_marked_answers": 0})
    def test_change_best_answer_no_permission(self):
        """api validates permission to change best answers"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to change this thread's marked answer "
                    'because it\'s in the "First category" category.'
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 2, "can_change_marked_answers": 1})
    def test_change_best_answer_not_starter(self):
        """api validates permission to change best answers"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to change this thread's marked answer "
                    "because you are not a thread starter."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

        # passing scenario is possible
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl(
        {
            "can_mark_best_answers": 2,
            "can_change_marked_answers": 1,
            "best_answer_change_time": 5,
        }
    )
    def test_change_best_answer_timelimit_out_of_time(self):
        """
        api validates permission for starter to change best answers within timelimit
        """
        best_answer = test.reply_thread(self.thread)

        self.thread.best_answer_marked_on = timezone.now() - timedelta(minutes=6)
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to change best answer that was marked "
                    "for more than 5 minutes."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl(
        {
            "can_mark_best_answers": 2,
            "can_change_marked_answers": 1,
            "best_answer_change_time": 5,
        }
    )
    def test_change_best_answer_timelimit(self):
        """
        api validates permission for starter to change best answers within timelimit
        """
        best_answer = test.reply_thread(self.thread)

        self.thread.best_answer_marked_on = timezone.now() - timedelta(minutes=1)
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl(
        {
            "can_mark_best_answers": 2,
            "can_change_marked_answers": 2,
            "can_protect_posts": False,
        }
    )
    def test_change_best_answer_protected_no_permission(self):
        """api validates permission to change protected best answers"""
        best_answer = test.reply_thread(self.thread)

        self.thread.best_answer_is_protected = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to change this thread's best answer "
                    "because a moderator has protected it."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl(
        {
            "can_mark_best_answers": 2,
            "can_change_marked_answers": 2,
            "can_protect_posts": True,
        }
    )
    def test_change_best_answer_protected(self):
        """api validates permission to change protected best answers"""
        best_answer = test.reply_thread(self.thread)

        self.thread.best_answer_is_protected = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_mark_best_answers": 2, "can_change_marked_answers": 2})
    def test_change_best_answer_post_validation(self):
        """api validates new post'"""
        best_answer = test.reply_thread(self.thread, is_hidden=True)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)


class ThreadUnmarkBestAnswerApiTests(ThreadPatchApiTestCase):
    def setUp(self):
        super().setUp()

        self.best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, self.best_answer)
        self.thread.save()

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 2})
    def test_unmark_best_answer(self):
        """api makes it possible to unmark best answer"""
        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": ["ok"],
                "best_answer": None,
                "best_answer_is_protected": False,
                "best_answer_marked_on": None,
                "best_answer_marked_by": None,
                "best_answer_marked_by_name": None,
                "best_answer_marked_by_slug": None,
            },
        )

        thread_json = self.get_thread_json()
        self.assertIsNone(thread_json["best_answer"])
        self.assertFalse(thread_json["best_answer_is_protected"])
        self.assertIsNone(thread_json["best_answer_marked_on"])
        self.assertIsNone(thread_json["best_answer_marked_by"])
        self.assertIsNone(thread_json["best_answer_marked_by_name"])
        self.assertIsNone(thread_json["best_answer_marked_by_slug"])

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 2})
    def test_unmark_best_answer_invalid_post_id(self):
        """api validates that post id is int"""
        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": "d7sd89a7d98sa"}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.id, "detail": ["A valid integer is required."]},
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 2})
    def test_unmark_best_answer_post_not_found(self):
        """api validates that post to unmark exists"""
        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id + 1}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"id": self.thread.id, "detail": ["NOT FOUND"]}
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 2})
    def test_unmark_best_answer_wrong_post(self):
        """api validates if post given to unmark is best answer"""
        best_answer = test.reply_thread(self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "This post can't be unmarked because it's not currently marked "
                    "as best answer."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 0})
    def test_unmark_best_answer_no_permission(self):
        """api validates if user has permission to unmark best answers"""
        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to unmark threads answers in "
                    'the "First category" category.'
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl({"can_mark_best_answers": 0, "can_change_marked_answers": 1})
    def test_unmark_best_answer_not_starter(self):
        """api validates if starter has permission to unmark best answers"""
        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to unmark this best answer because "
                    "you are not a thread starter."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

        # passing scenario is possible
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 1,
            "best_answer_change_time": 5,
        }
    )
    def test_unmark_best_answer_timelimit(self):
        """
        api validates if starter has permission to unmark best answer within time limit
        """
        self.thread.best_answer_marked_on = timezone.now() - timedelta(minutes=6)
        self.thread.starter = self.user
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to unmark best answer that was marked "
                    "for more than 5 minutes."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

        # passing scenario is possible
        self.thread.best_answer_marked_on = timezone.now() - timedelta(minutes=2)
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 2,
            "can_close_threads": False,
        }
    )
    def test_unmark_best_answer_closed_category_no_permission(self):
        """
        api validates if user has permission to unmark best answer in closed category
        """
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to unmark this best answer because "
                    'its category "First category" is closed.'
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 2,
            "can_close_threads": True,
        }
    )
    def test_unmark_best_answer_closed_category(self):
        """
        api validates if user has permission to unmark best answer in closed category
        """
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 2,
            "can_close_threads": False,
        }
    )
    def test_unmark_best_answer_closed_thread_no_permission(self):
        """
        api validates if user has permission to unmark best answer in closed thread
        """
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You can't unmark this thread's best answer because it's closed "
                    "and you don't have permission to open it."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 2,
            "can_close_threads": True,
        }
    )
    def test_unmark_best_answer_closed_thread(self):
        """
        api validates if user has permission to unmark best answer in closed thread
        """
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 2,
            "can_protect_posts": 0,
        }
    )
    def test_unmark_best_answer_protected_no_permission(self):
        """api validates permission to unmark protected best answers"""
        self.thread.best_answer_is_protected = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.id,
                "detail": [
                    "You don't have permission to unmark this thread's best answer "
                    "because a moderator has protected it."
                ],
            },
        )

        thread_json = self.get_thread_json()
        self.assertEqual(thread_json["best_answer"], self.best_answer.id)

    @patch_category_acl(
        {
            "can_mark_best_answers": 0,
            "can_change_marked_answers": 2,
            "can_protect_posts": 1,
        }
    )
    def test_unmark_best_answer_protected(self):
        """api validates permission to unmark protected best answers"""
        self.thread.best_answer_is_protected = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "best-answer", "value": self.best_answer.id}],
        )
        self.assertEqual(response.status_code, 200)



import importlib
import re
import sys
from pathlib import Path
from typing import Dict, List


def discover_plugins(plugins_path: str | None) -> List[str]:
    if not plugins_path:
        return []

    plugins_path_obj = Path(plugins_path)
    if not plugins_path_obj.is_dir():
        return []

    plugins_apps = discover_plugins_in_directory(plugins_path_obj)

    pip_install = plugins_path_obj / "pip-install.txt"
    if pip_install.is_file():
        plugins_apps += discover_plugins_in_pip_install(pip_install)

    return sorted(plugins_apps)


def discover_plugins_in_directory(plugins_path: Path) -> List[str]:
    plugins_apps: List[str] = []
    plugins_paths: Dict[str, str] = {}

    # First step: glob plugin Python packages
    for plugin_path in sorted(plugins_path.glob("*/*/misago_plugin.py")):
        plugin_package = plugin_path.parent

        # Skip plugins that are not valid Python packages
        plugin_package_init = plugin_package / "__init__.py"
        if not plugin_package_init.is_file():
            continue

        # Add plugin package name to Django apps for later import
        plugins_apps.append(plugin_package.name)

        # Store plugin path for later adding to sys.path
        plugins_paths[plugin_package.name] = str(plugin_package.parent)

    plugins_apps = sorted(plugins_apps)

    # Add unique plugins paths to Python path ordered by app name
    for plugin_app in plugins_apps:
        plugin_path = plugins_paths[plugin_app]
        if plugin_path not in sys.path:
            sys.path.append(plugin_path)

    return plugins_apps


def discover_plugins_in_pip_install(pip_install_path: Path) -> List[str]:
    pip_install_lines = read_pip_install_file(pip_install_path)
    return validate_modules_from_pip_install(pip_install_lines)


PIP_LINE_RE = re.compile("^[a-z0-9-_]+")


def read_pip_install_file(pip_install_path: Path) -> List[str]:
    clean_lines: List[str] = []
    with open(pip_install_path, "r") as fp:
        file_lines: List[str] = fp.readlines()
        for file_line in file_lines:
            clean_line = file_line.strip()
            if clean_line.startswith("#"):
                continue

            if "#" in clean_line:
                comment_start = clean_line.find("#")
                clean_line = clean_line[:comment_start].strip()

            clean_line_match = PIP_LINE_RE.match(clean_line)
            if not clean_line_match:
                continue

            clean_line = clean_line_match.group(0).replace("-", "_")
            if clean_line and clean_line not in clean_lines:
                clean_lines.append(clean_line)

    return clean_lines


def validate_modules_from_pip_install(pip_install_lines: List[str]) -> List[str]:
    valid_lines: List[str] = []
    for pip_install_line in pip_install_lines:
        try:
            module = importlib.import_module(pip_install_line)
            module_path = Path(module.__file__).parent
            if not module_path.is_dir():
                continue
            misago_plugin_path = module_path / "misago_plugin.py"
            if not misago_plugin_path.is_file():
                continue
            valid_lines.append(pip_install_line)
        except:
            pass
    return valid_lines



# Generated by Django 1.10.5 on 2017-03-04 22:06
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0007_auto_20170219_1639")]

    operations = [
        migrations.AddField(
            model_name="ban",
            name="registration_only",
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AlterField(
            model_name="ban",
            name="check_type",
            field=models.PositiveIntegerField(
                choices=[(0, "Username"), (1, "E-mail address"), (2, "IP address")],
                db_index=True,
                default=0,
            ),
        ),
    ]



from django.urls import reverse

from ...test import assert_contains


def test_admin_displays_own_error_page_on_404_error(admin_client):
    response = admin_client.get(reverse("misago:admin:index") + "404-not-found/")
    assert_contains(response, "Administration", 404)
    assert_contains(response, "Requested page could not be found", 404)



from .. import moderation, test
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..models import Post, Thread


class PostsModerationTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.all_categories()[:1][0]
        self.thread = test.post_thread(self.category)
        self.post = test.reply_thread(self.thread)

    def reload_thread(self):
        self.thread = Thread.objects.get(pk=self.thread.pk)

    def reload_post(self):
        self.post = Post.objects.get(pk=self.post.pk)

    def test_hide_original_post(self):
        """hide_post fails for first post in thread"""
        with self.assertRaises(moderation.ModerationError):
            moderation.hide_post(self.user, self.thread.first_post)

    def test_protect_post(self):
        """protect_post protects post"""
        self.assertFalse(self.post.is_protected)
        self.assertTrue(moderation.protect_post(self.user, self.post))

        self.reload_post()
        self.assertTrue(self.post.is_protected)

    def test_protect_protected_post(self):
        """protect_post fails to protect protected post gracefully"""
        self.post.is_protected = True
        self.assertFalse(moderation.protect_post(self.user, self.post))

    def test_unprotect_post(self):
        """unprotect_post releases post protection"""
        self.post.is_protected = True
        self.assertTrue(moderation.unprotect_post(self.user, self.post))

        self.reload_post()
        self.assertFalse(self.post.is_protected)

    def test_unprotect_protected_post(self):
        """unprotect_post fails to unprotect unprotected post gracefully"""
        self.assertFalse(moderation.unprotect_post(self.user, self.post))

    def test_hide_post(self):
        """hide_post hides post"""
        self.assertFalse(self.post.is_hidden)
        self.assertTrue(moderation.hide_post(self.user, self.post))

        self.reload_post()
        self.assertTrue(self.post.is_hidden)
        self.assertEqual(self.post.hidden_by, self.user)
        self.assertEqual(self.post.hidden_by_name, self.user.username)
        self.assertEqual(self.post.hidden_by_slug, self.user.slug)
        self.assertIsNotNone(self.post.hidden_on)

    def test_hide_hidden_post(self):
        """hide_post fails to hide hidden post gracefully"""
        self.post.is_hidden = True
        self.assertFalse(moderation.hide_post(self.user, self.post))

    def test_unhide_original_post(self):
        """unhide_post fails for first post in thread"""
        with self.assertRaises(moderation.ModerationError):
            moderation.unhide_post(self.user, self.thread.first_post)

    def test_unhide_post(self):
        """unhide_post reveals post"""
        self.post.is_hidden = True

        self.assertTrue(self.post.is_hidden)
        self.assertTrue(moderation.unhide_post(self.user, self.post))

        self.reload_post()
        self.assertFalse(self.post.is_hidden)

    def test_unhide_visible_post(self):
        """unhide_post fails to reveal visible post gracefully"""
        self.assertFalse(moderation.unhide_post(self.user, self.post))

    def test_delete_original_post(self):
        """delete_post fails for first post in thread"""
        with self.assertRaises(moderation.ModerationError):
            moderation.delete_post(self.user, self.thread.first_post)

    def test_delete_post(self):
        """delete_post deletes thread post"""
        self.assertTrue(moderation.delete_post(self.user, self.post))
        with self.assertRaises(Post.DoesNotExist):
            self.reload_post()

        self.thread.synchronize()
        self.assertEqual(self.thread.replies, 0)



from .versions import get_cache_versions


class assert_invalidates_cache:
    def __init__(self, cache):
        self.cache = cache

    def __enter__(self):
        self.versions = get_cache_versions()
        return self

    def __exit__(self, exc_type, *_):
        if exc_type:
            return False

        new_versions = get_cache_versions()
        for cache, version in new_versions.items():
            if cache == self.cache:
                message = "cache %s was not invalidated" % cache
                assert self.versions[cache] != version, message



from pathlib import Path

import pytest
from django.core.files.base import ContentFile
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_has_success_message
from ... import THEME_CACHE


@pytest.fixture
def delete_css(admin_client):
    def delete_assets(theme, assets):
        url = reverse("misago:admin:themes:delete-css", kwargs={"pk": theme.pk})
        return admin_client.post(url, {"item": [i.pk for i in assets]})

    return delete_assets


@pytest.fixture
def delete_media(admin_client):
    def delete_assets(theme, assets):
        url = reverse("misago:admin:themes:delete-media", kwargs={"pk": theme.pk})
        return admin_client.post(url, {"item": [i.pk for i in assets]})

    return delete_assets


def test_theme_css_can_be_deleted(theme, delete_css, css):
    delete_css(theme, [css])
    assert not theme.css.exists()


def test_deleting_css_also_deletes_css_source_files(theme, delete_css, css):
    delete_css(theme, [css])
    assert not Path(css.source_file.path).exists()


def test_deleting_css_also_deletes_css_build_files(theme, delete_css, css):
    css.build_file = ContentFile("body {}", name="test.css")
    css.build_hash = "abcdefgh"
    css.save()

    delete_css(theme, [css])
    assert not Path(css.build_file.path).exists()


def test_theme_css_link_can_be_deleted(theme, delete_css, css_link):
    delete_css(theme, [css_link])
    assert not theme.css.exists()


def test_multiple_theme_css_can_be_deleted_at_single_time(
    theme, delete_css, css, css_link
):
    delete_css(theme, [css, css_link])
    assert not theme.css.exists()


def test_theme_media_can_be_deleted(theme, delete_media, media):
    delete_media(theme, [media])
    assert not theme.media.exists()


def test_deleting_media_also_deletes_files(theme, delete_media, media):
    delete_media(theme, [media])
    assert not Path(media.file.path).exists()


def test_theme_images_can_be_deleted(theme, delete_media, image):
    delete_media(theme, [image])
    assert not theme.media.exists()


def test_deleting_image_also_deletes_files(theme, delete_media, image):
    delete_media(theme, [image])
    assert not Path(image.thumbnail.path).exists()


def test_multiple_theme_media_can_be_deleted_at_single_time(
    theme, delete_media, media, image
):
    delete_media(theme, [media, image])
    assert not theme.media.exists()


def test_success_message_is_set_after_css_is_deleted(theme, delete_css, css):
    response = delete_css(theme, [css])
    assert_has_success_message(response)


def test_success_message_is_set_after_media_is_deleted(theme, delete_media, media):
    response = delete_media(theme, [media])
    assert_has_success_message(response)


def test_selecting_no_css_to_delete_causes_no_errors(theme, delete_css, css):
    delete_css(theme, [])
    assert theme.css.exists()


def test_selecting_no_media_to_delete_causes_no_errors(theme, delete_media, media):
    delete_media(theme, [])
    assert theme.media.exists()


def test_selecting_invalid_css_id_to_delete_causes_no_errors(
    mocker, theme, delete_css, css
):
    delete_css(theme, [mocker.Mock(pk="str")])
    assert theme.css.exists()


def test_selecting_invalid_media_id_to_delete_causes_no_errors(
    mocker, theme, delete_media, media
):
    delete_media(theme, [mocker.Mock(pk="str")])
    assert theme.media.exists()


def test_selecting_nonexisting_css_id_to_delete_causes_no_errors(
    mocker, theme, delete_css, css
):
    delete_css(theme, [mocker.Mock(pk=css.pk + 1)])
    assert theme.css.exists()


def test_selecting_nonexisting_media_id_to_delete_causes_no_errors(
    mocker, theme, delete_media, media
):
    delete_media(theme, [mocker.Mock(pk=media.pk + 1)])
    assert theme.media.exists()


def test_other_theme_css_is_not_deleted(delete_css, theme, other_theme, css):
    delete_css(other_theme, [css])
    assert theme.css.exists()


def test_other_theme_media_is_not_deleted(delete_media, theme, other_theme, media):
    delete_media(other_theme, [media])
    assert theme.media.exists()


def test_deleting_css_invalidates_theme_cache(theme, delete_css, css):
    with assert_invalidates_cache(THEME_CACHE):
        delete_css(theme, [css])



from django.contrib.auth import logout
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404
from django.utils.translation import pgettext
from rest_framework.decorators import api_view
from rest_framework.response import Response

from .models import Agreement
from .utils import save_user_agreement_acceptance


@api_view(["POST"])
def submit_agreement(request, pk):
    agreement = get_object_or_404(Agreement, is_active=True, pk=pk)

    if agreement.id in request.user.agreements:
        raise PermissionDenied(
            pgettext("agreement submit", "You have already accepted this agreement.")
        )

    if request.data.get("accept") is True:
        save_user_agreement_acceptance(request.user, agreement, commit=True)
    elif request.data.get("accept") is False:
        if not request.user.is_staff:
            request.user.mark_for_delete()
            logout(request)
    else:
        raise PermissionDenied(
            pgettext("agreement submit", "You need to submit a valid choice.")
        )

    return Response({"detail": "ok"})



from django.urls import reverse

from ...users.test import AuthenticatedUserTestCase
from ..context_processors import legal_links
from ..models import Agreement


class MockRequest:
    def __init__(self, user):
        self.user = user
        self.frontend_context = {}

    def get_host(self):
        return "testhost.com"


class PrivacyPolicyTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        Agreement.objects.invalidate_cache()

    def tearDown(self):
        Agreement.objects.invalidate_cache()

    def test_context_processor_no_policy(self):
        """context processor has no TOS link"""
        context_dict = legal_links(MockRequest(self.user))
        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": None,
                "TERMS_OF_SERVICE_TITLE": None,
                "TERMS_OF_SERVICE_URL": None,
                "PRIVACY_POLICY_ID": None,
                "PRIVACY_POLICY_TITLE": None,
                "PRIVACY_POLICY_URL": None,
                "misago_agreement": None,
            },
        )

    def test_context_processor_misago_policy(self):
        """context processor has TOS link to Misago view"""
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, text="Lorem ipsum", is_active=True
        )

        context_dict = legal_links(MockRequest(self.user))

        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": None,
                "TERMS_OF_SERVICE_TITLE": None,
                "TERMS_OF_SERVICE_URL": None,
                "PRIVACY_POLICY_ID": agreement.id,
                "PRIVACY_POLICY_TITLE": "Privacy policy",
                "PRIVACY_POLICY_URL": reverse("misago:privacy-policy"),
                "misago_agreement": {
                    "type": "Privacy policy",
                    "title": "Privacy policy",
                    "link": None,
                    "text": "<p>Lorem ipsum</p>",
                },
            },
        )

    def test_context_processor_remote_policy(self):
        """context processor has TOS link to remote url"""
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, link="http://test.com", is_active=True
        )

        context_dict = legal_links(MockRequest(self.user))

        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": None,
                "TERMS_OF_SERVICE_TITLE": None,
                "TERMS_OF_SERVICE_URL": None,
                "PRIVACY_POLICY_ID": agreement.id,
                "PRIVACY_POLICY_TITLE": "Privacy policy",
                "PRIVACY_POLICY_URL": "http://test.com",
                "misago_agreement": {
                    "type": "Privacy policy",
                    "title": "Privacy policy",
                    "link": "http://test.com",
                    "text": None,
                },
            },
        )

        # set misago view too
        agreement.text = "Lorem ipsum"
        agreement.save()

        context_dict = legal_links(MockRequest(self.user))

        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": None,
                "TERMS_OF_SERVICE_TITLE": None,
                "TERMS_OF_SERVICE_URL": None,
                "PRIVACY_POLICY_ID": agreement.id,
                "PRIVACY_POLICY_TITLE": "Privacy policy",
                "PRIVACY_POLICY_URL": "http://test.com",
                "misago_agreement": {
                    "type": "Privacy policy",
                    "title": "Privacy policy",
                    "link": "http://test.com",
                    "text": "<p>Lorem ipsum</p>",
                },
            },
        )


class TermsOfServiceTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        Agreement.objects.invalidate_cache()

    def tearDown(self):
        Agreement.objects.invalidate_cache()

    def test_context_processor_no_tos(self):
        """context processor has no TOS link"""
        context_dict = legal_links(MockRequest(self.user))
        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": None,
                "TERMS_OF_SERVICE_TITLE": None,
                "TERMS_OF_SERVICE_URL": None,
                "PRIVACY_POLICY_ID": None,
                "PRIVACY_POLICY_TITLE": None,
                "PRIVACY_POLICY_URL": None,
                "misago_agreement": None,
            },
        )

    def test_context_processor_misago_tos(self):
        """context processor has TOS link to Misago view"""
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=True
        )

        context_dict = legal_links(MockRequest(self.user))

        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": agreement.id,
                "TERMS_OF_SERVICE_TITLE": "Terms of service",
                "TERMS_OF_SERVICE_URL": reverse("misago:terms-of-service"),
                "PRIVACY_POLICY_ID": None,
                "PRIVACY_POLICY_TITLE": None,
                "PRIVACY_POLICY_URL": None,
                "misago_agreement": {
                    "type": "Terms of service",
                    "title": "Terms of service",
                    "link": None,
                    "text": "<p>Lorem ipsum</p>",
                },
            },
        )

    def test_context_processor_remote_tos(self):
        """context processor has TOS link to remote url"""
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_TOS, link="http://test.com", is_active=True
        )

        context_dict = legal_links(MockRequest(self.user))

        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": agreement.id,
                "TERMS_OF_SERVICE_TITLE": "Terms of service",
                "TERMS_OF_SERVICE_URL": "http://test.com",
                "PRIVACY_POLICY_ID": None,
                "PRIVACY_POLICY_TITLE": None,
                "PRIVACY_POLICY_URL": None,
                "misago_agreement": {
                    "type": "Terms of service",
                    "title": "Terms of service",
                    "link": "http://test.com",
                    "text": None,
                },
            },
        )

        # set misago view too
        agreement.text = "Lorem ipsum"
        agreement.save()

        context_dict = legal_links(MockRequest(self.user))

        self.assertEqual(
            context_dict,
            {
                "TERMS_OF_SERVICE_ID": agreement.id,
                "TERMS_OF_SERVICE_TITLE": "Terms of service",
                "TERMS_OF_SERVICE_URL": "http://test.com",
                "PRIVACY_POLICY_ID": None,
                "PRIVACY_POLICY_TITLE": None,
                "PRIVACY_POLICY_URL": None,
                "misago_agreement": {
                    "type": "Terms of service",
                    "title": "Terms of service",
                    "link": "http://test.com",
                    "text": "<p>Lorem ipsum</p>",
                },
            },
        )



from django.http import HttpRequest, JsonResponse, Http404
from django.shortcuts import get_object_or_404
from rest_framework import serializers
from rest_framework.decorators import api_view

from ...categories.models import Category
from ...notifications.models import WatchedThread
from ...notifications.threads import ThreadNotifications, get_watched_thread
from ...threads.models import Thread
from ...threads.participants import make_thread_participants_aware
from ...threads.permissions import (
    allow_see_private_thread,
    allow_see_thread,
    allow_use_private_threads,
)
from ..decorators import require_auth


class ThreadWatchSerializer(serializers.Serializer):
    notifications = serializers.ChoiceField(
        choices=ThreadNotifications.choices,
    )


@api_view(["POST"])
@require_auth
def watch_thread(request: HttpRequest, thread_id: int) -> JsonResponse:
    thread = get_object_or_404(Thread.objects.select_related("category"), id=thread_id)
    if not Category.objects.root_category().has_child(thread.category):
        raise Http404()

    allow_see_thread(request.user_acl, thread)
    return watch_thread_shared_logic(request, thread)


@api_view(["POST"])
@require_auth
def watch_private_thread(request: HttpRequest, thread_id: int) -> JsonResponse:
    allow_use_private_threads(request.user_acl)

    thread = get_object_or_404(Thread, id=thread_id)
    if Category.objects.private_threads().id != thread.category_id:
        raise Http404()

    make_thread_participants_aware(request.user, thread)
    allow_see_private_thread(request.user_acl, thread)

    return watch_thread_shared_logic(request, thread)


def watch_thread_shared_logic(request: HttpRequest, thread: Thread) -> JsonResponse:
    serializer = ThreadWatchSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)
    notifications = serializer.data["notifications"]

    if not notifications:
        WatchedThread.objects.filter(user=request.user, thread=thread).delete()
        return JsonResponse(serializer.data)

    send_emails = notifications == ThreadNotifications.SITE_AND_EMAIL

    watched_thread = get_watched_thread(request.user, thread)
    if watched_thread:
        if watched_thread.send_emails != send_emails:
            watched_thread.send_emails = send_emails
            watched_thread.save(update_fields=["send_emails"])
    else:
        WatchedThread.objects.create(
            user=request.user,
            category_id=thread.category_id,
            thread=thread,
            send_emails=send_emails,
        )

    return JsonResponse(serializer.data)



import pytest
from django.forms import ValidationError

from ..validators import validate_css_name


def test_validation_fails_if_name_is_missing_css_extension():
    with pytest.raises(ValidationError):
        validate_css_name("filename")


def test_extension_validation_is_case_insensitive():
    validate_css_name("filename.CsS")


def test_validation_fails_if_name_starts_with_period():
    with pytest.raises(ValidationError):
        validate_css_name(".filename.css")


def test_validation_fails_if_name_contains_css_extension_only():
    with pytest.raises(ValidationError):
        validate_css_name(".css")


def test_validation_fails_if_name_contains_special_characters():
    with pytest.raises(ValidationError):
        validate_css_name("test().css")


def test_validation_fails_if_name_lacks_latin_characters_or_numbers():
    with pytest.raises(ValidationError):
        validate_css_name("_-.css")


def test_name_can_contain_underscores_scores_and_periods():
    validate_css_name("some_css-final2.dark.css")



# Generated by Django 1.11.13 on 2018-06-09 15:23
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0012_audittrail")]

    operations = [
        migrations.RemoveField(model_name="online", name="current_ip"),
        migrations.RemoveField(model_name="user", name="last_ip"),
        migrations.AlterField(
            model_name="user",
            name="joined_from_ip",
            field=models.GenericIPAddressField(blank=True, null=True),
        ),
    ]



import hashlib


def hash_email(email: str) -> str:
    return hashlib.md5(email.lower().encode()).hexdigest()


def slugify_username(username: str) -> str:
    return username.lower().strip().replace("_", "-")



import json
import os
from tempfile import TemporaryDirectory
from zipfile import BadZipFile, ZipFile

from django.core.files.uploadedfile import UploadedFile
from django.utils.translation import pgettext, pgettext_lazy

from ..models import Theme
from .css import create_css
from .forms import (
    ThemeCssUrlManifest,
    ThemeManifest,
    create_css_file_manifest,
    create_media_file_manifest,
)
from .media import create_media
from .tasks import build_theme_css, update_remote_css_size

INVALID_MANIFEST_ERROR = pgettext_lazy(
    "admin theme import",
    '"manifest.json" contained by ZIP file is not a valid theme manifest file.',
)


class ThemeImportError(BaseException):
    pass


class InvalidThemeManifest(ThemeImportError):
    def __init__(self):
        super().__init__(INVALID_MANIFEST_ERROR)


def import_theme(name, parent, zipfile):
    with TemporaryDirectory() as tmp_dir:
        extract_zipfile_to_tmp_dir(zipfile, tmp_dir)
        validate_zipfile_contains_single_directory(tmp_dir)
        theme_dir = os.path.join(tmp_dir, os.listdir(tmp_dir)[0])
        cleaned_manifest = clean_theme_contents(theme_dir)

        theme = create_theme_from_manifest(name, parent, cleaned_manifest)

        try:
            create_css_from_manifest(theme_dir, theme, cleaned_manifest["css"])
            create_media_from_manifest(theme_dir, theme, cleaned_manifest["media"])
        except Exception:
            theme.delete()
            raise InvalidThemeManifest()
        else:
            for css in theme.css.filter(url__isnull=False):
                update_remote_css_size.delay(css.pk)
            build_theme_css.delay(theme.pk)

        return theme


def extract_zipfile_to_tmp_dir(zipfile, tmp_dir):
    try:
        ZipFile(zipfile).extractall(tmp_dir)
    except BadZipFile:
        raise ThemeImportError(
            pgettext("admin theme import", "Uploaded ZIP file could not be extracted.")
        )


def validate_zipfile_contains_single_directory(tmp_dir):
    dir_contents = os.listdir(tmp_dir)
    if not dir_contents:
        raise ThemeImportError(
            pgettext("admin theme import", "Uploaded ZIP file is empty.")
        )
    if len(dir_contents) > 1:
        raise ThemeImportError(
            pgettext(
                "admin theme import",
                "Uploaded ZIP file should contain single directory.",
            )
        )
    if not os.path.isdir(os.path.join(tmp_dir, dir_contents[0])):
        raise ThemeImportError(
            pgettext(
                "admin theme import", "Uploaded ZIP file didn't contain a directory."
            )
        )


def clean_theme_contents(theme_dir):
    manifest = read_manifest(theme_dir)
    return clean_manifest(theme_dir, manifest)


def read_manifest(theme_dir):
    try:
        with open(os.path.join(theme_dir, "manifest.json")) as fp:
            return json.load(fp)
    except FileNotFoundError:
        raise ThemeImportError(
            pgettext(
                "admin theme import",
                'Uploaded ZIP file didn\'t contain a "manifest.json".',
            )
        )
    except json.decoder.JSONDecodeError:
        raise ThemeImportError(
            pgettext(
                "admin theme import",
                '"manifest.json" contained by ZIP file is not a valid JSON file.',
            )
        )


def clean_manifest(theme_dir, manifest):
    if not isinstance(manifest, dict):
        raise InvalidThemeManifest()

    form = ThemeManifest(manifest)
    if not form.is_valid():
        raise InvalidThemeManifest()

    cleaned_manifest = form.cleaned_data.copy()
    cleaned_manifest["css"] = clean_css_list(theme_dir, manifest.get("css"))
    cleaned_manifest["media"] = clean_media_list(theme_dir, manifest.get("media"))

    return cleaned_manifest


def clean_css_list(theme_dir, manifest):
    if not isinstance(manifest, list):
        raise InvalidThemeManifest()

    theme_css_dir = os.path.join(theme_dir, "css")
    if not os.path.isdir(theme_css_dir):
        raise InvalidThemeManifest()

    cleaned_data = []
    for item in manifest:
        cleaned_data.append(clean_css_list_item(theme_css_dir, item))
    return cleaned_data


def clean_css_list_item(theme_css_dir, item):
    if not isinstance(item, dict):
        raise InvalidThemeManifest()

    if item.get("url"):
        return clean_css_url(item)
    if item.get("path"):
        return clean_css_file(theme_css_dir, item)

    raise InvalidThemeManifest()


def clean_css_url(data):
    form = ThemeCssUrlManifest(data)
    if not form.is_valid():
        raise InvalidThemeManifest()
    return form.cleaned_data


def clean_css_file(theme_css_dir, data):
    file_manifest = create_css_file_manifest(theme_css_dir)

    if data.get("path"):
        data["path"] = os.path.join(theme_css_dir, str(data["path"]))

    form = file_manifest(data)
    if not form.is_valid():
        raise InvalidThemeManifest()
    return form.cleaned_data


def clean_media_list(theme_dir, manifest):
    if not isinstance(manifest, list):
        raise InvalidThemeManifest()

    theme_media_dir = os.path.join(theme_dir, "media")
    if not os.path.isdir(theme_media_dir):
        raise InvalidThemeManifest()

    cleaned_data = []
    for item in manifest:
        cleaned_data.append(clean_media_list_item(theme_media_dir, item))
    return cleaned_data


def clean_media_list_item(theme_media_dir, data):
    if not isinstance(data, dict):
        raise InvalidThemeManifest()

    file_manifest = create_media_file_manifest(theme_media_dir)

    if data.get("path"):
        data["path"] = os.path.join(theme_media_dir, str(data["path"]))

    form = file_manifest(data)
    if not form.is_valid():
        raise InvalidThemeManifest()

    return form.cleaned_data


def create_theme_from_manifest(name, parent, cleaned_data):
    return Theme.objects.create(
        name=name or cleaned_data["name"],
        parent=parent,
        version=cleaned_data.get("version") or None,
        author=cleaned_data.get("author") or None,
        url=cleaned_data.get("url") or None,
    )


def create_css_from_manifest(tmp_dir, theme, manifest):
    for item in manifest:
        if "url" in item:
            theme.css.create(**item, order=theme.css.count())
        else:
            with open(item["path"], "rb") as fp:
                file_obj = UploadedFile(
                    fp,
                    name=item["name"],
                    content_type="text/css",
                    size=os.path.getsize(item["path"]),
                )
                create_css(theme, file_obj)


def create_media_from_manifest(tmp_dir, theme, manifest):
    for item in manifest:
        with open(item["path"], "rb") as fp:
            file_obj = UploadedFile(
                fp,
                name=item["name"],
                content_type=item["type"],
                size=os.path.getsize(item["path"]),
            )
            create_media(theme, file_obj)



from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext
from rest_framework.response import Response

from ...models import Thread
from ...moderation import threads as moderation
from ...serializers import SplitPostsSerializer


def posts_split_endpoint(request, thread):
    if not thread.acl["can_move_posts"]:
        raise PermissionDenied(
            pgettext("posts api", "You can't split posts from this thread.")
        )

    serializer = SplitPostsSerializer(
        data=request.data,
        context={
            "settings": request.settings,
            "thread": thread,
            "user_acl": request.user_acl,
        },
    )

    if not serializer.is_valid():
        if "posts" in serializer.errors:
            # Fix for KeyError - errors[0]
            errors = serializer.errors["posts"]
            try:
                errors = {"detail": errors[0]}
            except KeyError:
                if isinstance(errors, dict):
                    errors = {"detail": list(errors.values())[0][0]}
        else:
            errors = serializer.errors

        return Response(errors, status=400)

    split_posts_to_new_thread(request, thread, serializer.validated_data)

    return Response({})


def split_posts_to_new_thread(request, thread, validated_data):
    new_thread = Thread(
        category=validated_data["category"],
        started_on=thread.started_on,
        last_post_on=thread.last_post_on,
    )

    new_thread.set_title(validated_data["title"])
    new_thread.save()

    for post in validated_data["posts"]:
        post.move(new_thread)
        post.save()

    thread.synchronize()
    thread.save()

    new_thread.synchronize()
    new_thread.save()

    if validated_data.get("weight") == Thread.WEIGHT_GLOBAL:
        moderation.pin_thread_globally(request, new_thread)
    elif validated_data.get("weight"):
        moderation.pin_thread_locally(request, new_thread)
    if validated_data.get("is_hidden", False):
        moderation.hide_thread(request, new_thread)
    if validated_data.get("is_closed", False):
        moderation.close_thread(request, new_thread)

    thread.category.synchronize()
    thread.category.save()

    if new_thread.category != thread.category:
        new_thread.category.synchronize()
        new_thread.category.save()



import os
from hashlib import md5
from io import BytesIO

from django.core.files import File
from django.core.files.base import ContentFile
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from PIL import Image

from ...conf import settings
from ...core.utils import slugify
from ...plugins.models import PluginDataModel


def upload_to(instance, filename):
    # pylint: disable=undefined-loop-variable
    spread_path = md5(str(instance.secret[:16]).encode()).hexdigest()
    secret = Attachment.generate_new_secret()

    filename_lowered = filename.lower().strip()
    for extension in instance.filetype.extensions_list:
        if filename_lowered.endswith(extension):
            break

    filename_clean = ".".join(
        (slugify(filename[: (len(extension) + 1) * -1])[:16], extension)
    )

    return os.path.join(
        "attachments", spread_path[:2], spread_path[2:4], secret, filename_clean
    )


class Attachment(PluginDataModel):
    secret = models.CharField(max_length=64)
    filetype = models.ForeignKey("AttachmentType", on_delete=models.CASCADE)
    post = models.ForeignKey("Post", blank=True, null=True, on_delete=models.SET_NULL)

    uploaded_on = models.DateTimeField(default=timezone.now, db_index=True)

    uploader = models.ForeignKey(
        settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL
    )
    uploader_name = models.CharField(max_length=255)
    uploader_slug = models.CharField(max_length=255, db_index=True)

    filename = models.CharField(max_length=255, db_index=True)
    size = models.PositiveIntegerField(default=0, db_index=True)

    thumbnail = models.ImageField(
        max_length=255, blank=True, null=True, upload_to=upload_to
    )
    image = models.ImageField(
        max_length=255, blank=True, null=True, upload_to=upload_to
    )
    file = models.FileField(max_length=255, blank=True, null=True, upload_to=upload_to)

    def __str__(self):
        return self.filename

    def delete(self, *args, **kwargs):
        self.delete_files()
        return super().delete(*args, **kwargs)

    def delete_files(self):
        if self.thumbnail:
            self.thumbnail.delete(save=False)
        if self.image:
            self.image.delete(save=False)
        if self.file:
            self.file.delete(save=False)

    @classmethod
    def generate_new_secret(cls):
        return get_random_string(settings.MISAGO_ATTACHMENT_SECRET_LENGTH)

    @property
    def is_image(self):
        return bool(self.image)

    @property
    def is_file(self):
        return not self.is_image

    def get_absolute_url(self):
        return reverse(
            "misago:attachment", kwargs={"pk": self.pk, "secret": self.secret}
        )

    def get_thumbnail_url(self):
        if self.thumbnail:
            return reverse(
                "misago:attachment-thumbnail",
                kwargs={"pk": self.pk, "secret": self.secret},
            )

    def set_file(self, upload):
        self.file = File(upload, upload.name)

    def set_image(self, upload):
        fileformat = self.filetype.extensions_list[0]

        self.image = File(upload, upload.name)

        thumbnail = Image.open(upload)
        downscale_image = (
            thumbnail.size[0] > settings.MISAGO_ATTACHMENT_IMAGE_SIZE_LIMIT[0]
            or thumbnail.size[1] > settings.MISAGO_ATTACHMENT_IMAGE_SIZE_LIMIT[1]
        )
        strip_animation = fileformat == "gif"

        thumb_stream = BytesIO()
        if downscale_image:
            thumbnail.thumbnail(settings.MISAGO_ATTACHMENT_IMAGE_SIZE_LIMIT)
            if fileformat == "jpg":
                # normalize jpg to jpeg for Pillow
                thumbnail.save(thumb_stream, "jpeg")
            else:
                thumbnail.save(thumb_stream, "png")
        elif strip_animation:
            thumbnail.save(thumb_stream, "png")

        if downscale_image or strip_animation:
            self.thumbnail = ContentFile(thumb_stream.getvalue(), upload.name)



from datetime import timedelta

import pytest
from django.urls import reverse
from django.utils import timezone

from ....admin.test import AdminTestCase
from ....cache.test import assert_invalidates_cache
from ....test import assert_contains
from ... import BANS_CACHE
from ...models import Ban


@pytest.fixture
def admin_link(admin_client):
    response = admin_client.get(reverse("misago:admin:users:bans:index"))
    return response["location"]


@pytest.fixture
def ban(db):
    return Ban.objects.create(banned_value="banned_username")


def test_link_is_registered_in_admin_nav(admin_client):
    response = admin_client.get(reverse("misago:admin:users:index"))
    response = admin_client.get(response["location"])
    assert_contains(response, reverse("misago:admin:users:bans:index"))


def test_list_renders_empty(admin_client, admin_link):
    response = admin_client.get(admin_link)
    assert response.status_code == 200


def test_list_renders_with_item(admin_client, admin_link, ban):
    response = admin_client.get(admin_link)
    assert_contains(response, ban.banned_value)


def test_bans_can_be_mass_deleted(admin_client, admin_link):
    bans_ids = [Ban.objects.create(banned_value="ban_%s" % i).id for i in range(10)]

    admin_client.post(admin_link, data={"action": "delete", "selected_items": bans_ids})

    assert Ban.objects.count() == 0


def test_mass_deleting_bans_invalidates_bans_cache(admin_client, admin_link, ban):
    with assert_invalidates_cache(BANS_CACHE):
        admin_client.post(
            admin_link, data={"action": "delete", "selected_items": [ban.id]}
        )


def test_ban_can_be_deleted(admin_client, ban):
    admin_client.post(reverse("misago:admin:users:bans:delete", kwargs={"pk": ban.pk}))

    with pytest.raises(Ban.DoesNotExist):
        ban.refresh_from_db()


def test_deleting_ban_invalidates_bans_cache(admin_client, ban):
    with assert_invalidates_cache(BANS_CACHE):
        admin_client.post(
            reverse("misago:admin:users:bans:delete", kwargs={"pk": ban.pk})
        )


def test_new_ban_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:users:bans:new"))
    assert response.status_code == 200


def test_new_ban_can_be_created(admin_client):
    test_date = timezone.now() + timedelta(days=180)

    response = admin_client.post(
        reverse("misago:admin:users:bans:new"),
        data={
            "check_type": Ban.EMAIL,
            "banned_value": "test@test.com",
            "user_message": "Lorem ipsum dolor met",
            "staff_message": "Sit amet elit",
            "expires_on": test_date.isoformat(),
        },
    )

    ban = Ban.objects.get()
    assert ban.check_type == Ban.EMAIL
    assert ban.banned_value == "test@test.com"
    assert ban.user_message == "Lorem ipsum dolor met"
    assert ban.staff_message == "Sit amet elit"
    assert ban.expires_on.isoformat() == test_date.isoformat()


def test_new_ban_creation_invalidates_bans_cache(admin_client):
    with assert_invalidates_cache(BANS_CACHE):
        admin_client.post(
            reverse("misago:admin:users:bans:new"),
            data={"check_type": Ban.EMAIL, "banned_value": "test@test.com"},
        )


def test_edit_ban_form_renders(admin_client, ban):
    response = admin_client.get(
        reverse("misago:admin:users:bans:edit", kwargs={"pk": ban.pk})
    )
    assert response.status_code == 200


def test_ban_can_be_edited(admin_client, ban):
    test_date = timezone.now() + timedelta(days=180)

    response = admin_client.post(
        reverse("misago:admin:users:bans:edit", kwargs={"pk": ban.pk}),
        data={
            "check_type": Ban.EMAIL,
            "banned_value": "test@test.com",
            "user_message": "Lorem ipsum dolor met",
            "staff_message": "Sit amet elit",
            "expires_on": test_date.isoformat(),
        },
    )

    ban.refresh_from_db()
    assert ban.check_type == Ban.EMAIL
    assert ban.banned_value == "test@test.com"
    assert ban.user_message == "Lorem ipsum dolor met"
    assert ban.staff_message == "Sit amet elit"
    assert ban.expires_on.isoformat() == test_date.isoformat()


def test_ban_edition_invalidates_bans_cache(admin_client, ban):
    with assert_invalidates_cache(BANS_CACHE):
        admin_client.post(
            reverse("misago:admin:users:bans:edit", kwargs={"pk": ban.pk}),
            data={"check_type": Ban.EMAIL, "banned_value": "test@test.com"},
        )



from django.contrib.postgres.search import SearchQuery, SearchRank, SearchVector
from django.utils.translation import pgettext_lazy

from ..conf import settings
from ..core.shortcuts import paginate, pagination_dict
from ..search import SearchProvider
from .filtersearch import filter_search
from .models import Post, Thread
from .permissions import exclude_invisible_threads
from .serializers import FeedSerializer
from .utils import add_categories_to_items
from .viewmodels import ThreadsRootCategory


class SearchThreads(SearchProvider):
    name = pgettext_lazy("search provider", "Threads")
    icon = "forum"
    url = "threads"

    def search(self, query, page=1):
        root_category = ThreadsRootCategory(self.request)
        threads_categories = [root_category.unwrap()] + root_category.subcategories

        if len(query) > 1:
            visible_threads = exclude_invisible_threads(
                self.request.user_acl, threads_categories, Thread.objects
            )
            results = search_threads(self.request, query, visible_threads)
        else:
            results = []

        list_page = paginate(
            results,
            page,
            self.request.settings.posts_per_page,
            self.request.settings.posts_per_page_orphans,
            allow_explicit_first_page=True,
        )
        paginator = pagination_dict(list_page)

        posts = []
        threads = []
        if paginator["count"]:
            posts = list(
                list_page.object_list.select_related("thread", "poster", "poster__rank")
            )

            threads = []
            for post in posts:
                threads.append(post.thread)

            add_categories_to_items(
                root_category.unwrap(), threads_categories, posts + threads
            )

        results = {
            "results": FeedSerializer(
                posts,
                many=True,
                context={
                    "settings": self.request.settings,
                    "user": self.request.user,
                },
            ).data
        }
        results.update(paginator)

        return results


def search_threads(request, query, visible_threads):
    max_hits = request.settings.posts_per_page * 5
    clean_query = filter_search(query)

    if not clean_query:
        # Short-circuit search due to empty cleaned query
        return Post.objects.none()

    search_query = SearchQuery(clean_query, config=settings.MISAGO_SEARCH_CONFIG)
    search_vector = SearchVector(
        "search_document", config=settings.MISAGO_SEARCH_CONFIG
    )

    queryset = Post.objects.filter(
        is_event=False,
        is_hidden=False,
        is_unapproved=False,
        thread_id__in=visible_threads.values("id"),
        search_vector=search_query,
    )

    if queryset[: max_hits + 1].count() > max_hits:
        queryset = queryset.order_by("-id")[:max_hits]

    return (
        Post.objects.filter(id__in=queryset.values("id"))
        .annotate(rank=SearchRank(search_vector, search_query))
        .order_by("-rank", "-id")
    )



from django.core.cache import cache

from . import SOCIALAUTH_CACHE
from ..cache.versions import invalidate_cache


def get_socialauth_cache(cache_versions):
    key = get_cache_key(cache_versions)
    return cache.get(key)


def set_socialauth_cache(cache_versions, socialauth):
    key = get_cache_key(cache_versions)
    cache.set(key, socialauth)


def get_cache_key(cache_versions):
    return "%s_%s" % (SOCIALAUTH_CACHE, cache_versions[SOCIALAUTH_CACHE])


def clear_socialauth_cache():
    invalidate_cache(SOCIALAUTH_CACHE)



from django.template import Context, Template
from django.utils.safestring import mark_safe

from ..outlets import PluginOutlet, append_outlet_action, prepend_outlet_action


def strong_action(request, context):
    body = context.get("value") or "none"
    return mark_safe(f"<strong>{body}</strong>")


def em_action(request, context):
    body = context.get("value") or "none"
    return mark_safe(f"<em>{body}</em>")


def test_empty_outlet_renders_nothing(patch_outlets, render_outlet_template, snapshot):
    html = render_outlet_template()
    assert snapshot == html


def test_outlet_renders_appended_plugin(
    patch_outlets, render_outlet_template, snapshot
):
    append_outlet_action(PluginOutlet.TEST, strong_action)
    html = render_outlet_template()

    assert snapshot == html


def test_outlet_renders_prepended_plugin(
    patch_outlets, render_outlet_template, snapshot
):
    prepend_outlet_action(PluginOutlet.TEST, strong_action)
    html = render_outlet_template()

    assert snapshot == html


def test_outlet_renders_multiple_plugins(
    patch_outlets, render_outlet_template, snapshot
):
    append_outlet_action(PluginOutlet.TEST, strong_action)
    prepend_outlet_action(PluginOutlet.TEST, em_action)
    prepend_outlet_action(PluginOutlet.TEST, strong_action)
    html = render_outlet_template()

    assert snapshot == html


def test_outlet_renders_plugins_with_context(
    patch_outlets, render_outlet_template, snapshot
):
    append_outlet_action(PluginOutlet.TEST, strong_action)
    prepend_outlet_action(PluginOutlet.TEST, em_action)
    prepend_outlet_action(PluginOutlet.TEST, strong_action)
    html = render_outlet_template({"value": "context"})

    assert snapshot == html


def render_hasplugins_template(context: dict | None = None):
    template = Template(
        """
        {% load misago_plugins %}
        <div>{% hasplugins TEST %}plugins{% endhasplugins %}</div>
        """
    )

    return template.render(Context(context or {})).strip()


def test_hasplugins_tag_renders_nothing_if_no_plugins_exist(patch_outlets, snapshot):
    html = render_hasplugins_template()

    assert snapshot == html


def test_hasplugins_tag_renders_value_if_plugins_exist(patch_outlets, snapshot):
    append_outlet_action(PluginOutlet.TEST, strong_action)
    html = render_hasplugins_template()

    assert snapshot == html


def render_haspluginselse_template(context: dict | None = None):
    template = Template(
        """
        {% load misago_plugins %}
        <div>{% hasplugins TEST %}plugins{% else %}noplugins{% endhasplugins %}</div>
        """
    )

    return template.render(Context(context or {})).strip()


def test_hasplugins_else_tag_renders_else_if_no_plugins_exist(patch_outlets, snapshot):
    html = render_haspluginselse_template()

    assert snapshot == html


def test_hasplugins_else_tag_renders_value_if_plugins_exist(patch_outlets, snapshot):
    append_outlet_action(PluginOutlet.TEST, strong_action)
    html = render_haspluginselse_template()

    assert snapshot == html



from unittest.mock import Mock

import pytest

from ..templatetags.misago_absoluteurl import absoluteurl


@pytest.fixture
def context():
    return {"settings": Mock(forum_address="http://test.com/")}


def test_path_is_prefixed_with_forum_address(context):
    assert absoluteurl(context, "/path/") == "http://test.com/path/"


def test_link_is_reversed_and_prefixed_with_forum_address(context):
    assert absoluteurl(context, "misago:index") == "http://test.com/"


def test_absolute_url_is_not_changed(context):
    url = "https://github.com/rafalp/Misago/issues/1067"
    assert absoluteurl(context, url) == url



from . import PostingEndpoint, PostingMiddleware
from ....categories import PRIVATE_THREADS_ROOT_NAME
from ...participants import set_users_unread_private_threads_sync


class SyncPrivateThreadsMiddleware(PostingMiddleware):
    """middleware that sets private thread participants to sync unread threads"""

    def use_this_middleware(self):
        if self.mode == PostingEndpoint.REPLY:
            return self.thread.thread_type.root_name == PRIVATE_THREADS_ROOT_NAME
        return False

    def post_save(self, serializer):
        set_users_unread_private_threads_sync(
            participants=self.thread.participants_list,
            exclude_user=self.user,
        )



from typing import Any, Dict, Optional

from django.http import HttpRequest
from django.urls import reverse, NoReverseMatch
from django.utils.translation import get_language
from django.shortcuts import render as dj_render

from ...conf import settings
from ...core.momentjs import get_locale_url
from .. import site
from ..auth import is_admin_authorized, update_admin_authorization
from ..momentjs import MISAGO_ADMIN_MOMENT_LOCALES
from .auth import login


def get_protected_namespace(request):
    for namespace in settings.MISAGO_ADMIN_NAMESPACES:
        try:
            admin_path = reverse("%s:index" % namespace)
            if request.path.startswith(admin_path):
                return namespace
        except NoReverseMatch:
            pass


def render(
    request: HttpRequest,
    template: str,
    context: Optional[Dict[str, Any]] = None,
    error_page: bool = False,
):
    context = context or {}

    navigation = site.visible_branches(request)
    sections = navigation[0]

    try:
        actions = navigation[1]
    except IndexError:
        actions = []

    context.update({"sections": sections, "actions": actions})

    if error_page:
        # admittedly haxy solution for displaying navs on error pages
        context["actions"] = []
        context["pages"] = []
        for item in sections:
            item["is_active"] = False
    else:
        context["active_section"] = None
        for item in sections:
            if item["is_active"]:
                context["active_section"] = item
                break

        context["active_link"] = None
        for nav in navigation:
            for item in nav:
                if item["is_active"]:
                    context["active_link"] = item
                    break

    context["ADMIN_MOMENTJS_LOCALE_URL"] = get_admin_moment_locale_url()

    return dj_render(request, template, context)


def get_admin_moment_locale_url():
    return get_locale_url(
        get_language(),
        static_path_template="misago/admin/momentjs/%s.js",
        locales=MISAGO_ADMIN_MOMENT_LOCALES,
    )


# Decorator for views
def protected_admin_view(f):
    def decorator(request, *args, **kwargs):
        protected_view = get_protected_namespace(request)
        if protected_view:
            if is_admin_authorized(request):
                update_admin_authorization(request)
                return f(request, *args, **kwargs)
            request.admin_namespace = protected_view
            return login(request)
        return f(request, *args, **kwargs)

    return decorator



import random

from ..categories.models import Category, RoleCategoryACL
from .colors import COLORS


def fake_category(fake, parent, copy_acl_from=None):
    category = Category()
    category.set_name(fake_category_name(fake))

    if random.randint(1, 100) > 50:
        category.description = fake_category_description(fake)

    if random.randint(1, 100) > 15:
        category.color = random.choice(COLORS)

    category.insert_at(parent, position="last-child", save=True)

    if copy_acl_from:
        copy_acl_to_fake_category(copy_acl_from, category)

    return category


def fake_closed_category(fake, parent, copy_acl_from=None):
    category = fake_category(fake, parent, copy_acl_from)
    category.is_closed = True
    category.save(update_fields=["is_closed"])

    return category


def copy_acl_to_fake_category(source, category):
    copied_acls = []
    for acl in source.category_role_set.all():
        copied_acls.append(
            RoleCategoryACL(
                role_id=acl.role_id,
                category=category,
                category_role_id=acl.category_role_id,
            )
        )

    if copied_acls:
        RoleCategoryACL.objects.bulk_create(copied_acls)


def fake_category_name(fake):
    if random.randint(1, 100) > 75:
        return fake.catch_phrase().title()
    return fake.street_name()


def fake_category_description(fake):
    if random.randint(1, 100) > 80:
        return "\r\n".join(fake.paragraphs())
    return fake.paragraph()



from datetime import timedelta

from django.core.management.base import BaseCommand
from django.utils import timezone

from ....conf.shortcuts import get_dynamic_settings
from ...models import Notification


class Command(BaseCommand):
    help = "Deletes old notifications"

    def handle(self, *args, **options):
        settings = get_dynamic_settings()
        cutoff_date = timezone.now() - timedelta(
            days=settings.delete_notifications_older_than
        )
        queryset = Notification.objects.filter(created_at__lt=cutoff_date)
        deleted_count = queryset.count()

        if deleted_count:
            queryset.delete()
            message = "\n\nDeleted %s old notifications." % deleted_count
        else:
            message = "\n\nNo old notifications have been deleted."

        self.stdout.write(message)



from hashlib import md5
from time import time

from django.contrib import auth as dj_auth
from django.contrib import messages
from django.utils.translation import pgettext

from ..conf import settings

TOKEN_KEY = "misago_admin_session_token"
UPDATED_KEY = "misago_admin_session_updated"


# Admin session state controls
def is_admin_authorized(request):
    if request.user.is_anonymous:
        return False

    if not request.user.is_staff:
        return False

    admin_token = request.session.get(TOKEN_KEY)
    if not admin_token == make_user_admin_token(request.user):
        return False

    updated = request.session.get(UPDATED_KEY, 0)
    if updated < time() - (settings.MISAGO_ADMIN_SESSION_EXPIRATION * 60):
        if updated:
            request.session.pop(UPDATED_KEY, None)
            messages.info(
                request, pgettext("admin auth", "Your admin session has expired.")
            )
        return False

    return True


def authorize_admin(request):
    request.session[TOKEN_KEY] = make_user_admin_token(request.user)
    request.session[UPDATED_KEY] = int(time())


def update_admin_authorization(request):
    request.session[UPDATED_KEY] = int(time())


def remove_admin_authorization(request):
    request.session.pop(TOKEN_KEY, None)
    request.session.pop(UPDATED_KEY, None)


def make_user_admin_token(user):
    formula = (str(user.pk), user.email, user.password, settings.SECRET_KEY)
    return md5(":".join(formula).encode()).hexdigest()


# Login/logout exposed
login = dj_auth.login
logout = dj_auth.logout


# Register signal for logout to make sure eventual admin session is closed
def django_login_handler(sender, **kwargs):
    request, user = kwargs["request"], kwargs["user"]
    try:
        admin_namespace = request.admin_namespace
    except AttributeError:
        admin_namespace = False
    if admin_namespace and user.is_staff:
        authorize_admin(request)


dj_auth.signals.user_logged_in.connect(django_login_handler)


def django_logout_handler(sender, **kwargs):
    remove_admin_authorization(kwargs["request"])


dj_auth.signals.user_logged_out.connect(django_logout_handler)



from unittest.mock import Mock

import pytest

from ....notifications.models import Notification
from ..pagination import PaginationError, paginate_queryset


def test_pagination_raises_error_if_after_is_not_a_number():
    request = Mock(GET={"after": "str"})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_before_is_not_a_number():
    request = Mock(GET={"before": "str"})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_limit_is_not_a_number():
    request = Mock(GET={"limit": "str"})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_after_is_zero():
    request = Mock(GET={"after": 0})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_before_is_zero():
    request = Mock(GET={"before": 0})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_limit_is_zero():
    request = Mock(GET={"limit": 0})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_after_is_negative():
    request = Mock(GET={"after": -1})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_before_is_negative():
    request = Mock(GET={"before": -1})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_limit_is_negative():
    request = Mock(GET={"limit": -1})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "must be a positive integer" in str(excinfo)


def test_pagination_raises_error_if_after_and_before_are_both_set():
    request = Mock(GET={"after": 1, "before": 5})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "'after' and 'before' can't be used at same time" in str(excinfo)


def test_pagination_raises_error_if_limit_exceeds_max_value():
    request = Mock(GET={"limit": 200})

    with pytest.raises(PaginationError) as excinfo:
        paginate_queryset(request, Notification.objects, "id", 100)

    assert "can't be greater than '100'" in str(excinfo)



from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied, ValidationError
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.translation import npgettext, pgettext
from rest_framework import serializers
from rest_framework.response import Response

from ....acl import useracl
from ....acl.objectacl import add_acl_to_obj
from ....categories.models import Category
from ....categories.permissions import allow_browse_category, allow_see_category
from ....categories.serializers import CategorySerializer
from ....conf import settings
from ....core.apipatch import ApiPatch
from ....core.shortcuts import get_int_or_404
from ....users.utils import slugify_username
from ...moderation import threads as moderation
from ...participants import (
    add_participant,
    change_owner,
    make_participants_aware,
    remove_participant,
)
from ...permissions import (
    allow_add_participant,
    allow_add_participants,
    allow_approve_thread,
    allow_change_best_answer,
    allow_change_owner,
    allow_edit_thread,
    allow_hide_thread,
    allow_mark_as_best_answer,
    allow_mark_best_answer,
    allow_move_thread,
    allow_pin_thread,
    allow_remove_participant,
    allow_see_post,
    allow_start_thread,
    allow_unhide_thread,
    allow_unmark_best_answer,
)
from ...serializers import ThreadParticipantSerializer
from ...validators import validate_thread_title

User = get_user_model()

thread_patch_dispatcher = ApiPatch()


def patch_acl(request, thread, value):
    """useful little op that updates thread acl to current state"""
    if value:
        add_acl_to_obj(request.user_acl, thread)
        return {"acl": thread.acl}
    return {"acl": None}


thread_patch_dispatcher.add("acl", patch_acl)


def patch_title(request, thread, value):
    try:
        value_cleaned = str(value).strip()
    except (TypeError, ValueError):
        raise PermissionDenied(pgettext("threads api", "Not a valid string."))

    try:
        validate_thread_title(request.settings, value_cleaned)
    except ValidationError as e:
        raise PermissionDenied(e.args[0])

    allow_edit_thread(request.user_acl, thread)

    moderation.change_thread_title(request, thread, value_cleaned)
    return {"title": thread.title}


thread_patch_dispatcher.replace("title", patch_title)


def patch_weight(request, thread, value):
    allow_pin_thread(request.user_acl, thread)

    if not thread.acl.get("can_pin_globally") and thread.weight == 2:
        raise PermissionDenied(
            pgettext(
                "threads api",
                "You can't change globally pinned threads weights in this category.",
            )
        )

    if value == 2:
        if thread.acl.get("can_pin_globally"):
            moderation.pin_thread_globally(request, thread)
        else:
            raise PermissionDenied(
                pgettext(
                    "threads api", "You can't pin threads globally in this category."
                )
            )
    elif value == 1:
        moderation.pin_thread_locally(request, thread)
    elif value == 0:
        moderation.unpin_thread(request, thread)

    return {"weight": thread.weight}


thread_patch_dispatcher.replace("weight", patch_weight)


def patch_move(request, thread, value):
    allow_move_thread(request.user_acl, thread)

    category_pk = get_int_or_404(value)
    new_category = get_object_or_404(
        Category.objects.all_categories().select_related("parent"), pk=category_pk
    )

    add_acl_to_obj(request.user_acl, new_category)
    allow_see_category(request.user_acl, new_category)
    allow_browse_category(request.user_acl, new_category)
    allow_start_thread(request.user_acl, new_category)

    if new_category == thread.category:
        raise PermissionDenied(
            pgettext(
                "threads api", "You can't move thread to the category it's already in."
            )
        )

    moderation.move_thread(request, thread, new_category)

    return {
        "category": CategorySerializer(
            new_category,
            context={
                "settings": request.settings,
            },
        ).data,
    }


thread_patch_dispatcher.replace("category", patch_move)


def patch_flatten_categories(request, thread, value):
    try:
        return {"category": thread.category_id}
    except AttributeError:
        return {"category": thread.category_id}


thread_patch_dispatcher.replace("flatten-categories", patch_flatten_categories)


def patch_is_unapproved(request, thread, value):
    allow_approve_thread(request.user_acl, thread)

    if value:
        raise PermissionDenied(
            pgettext("threads api", "Content approval can't be reversed.")
        )

    moderation.approve_thread(request, thread)

    return {
        "is_unapproved": thread.is_unapproved,
        "has_unapproved_posts": thread.has_unapproved_posts,
    }


thread_patch_dispatcher.replace("is-unapproved", patch_is_unapproved)


def patch_is_closed(request, thread, value):
    if thread.acl.get("can_close"):
        if value:
            moderation.close_thread(request, thread)
        else:
            moderation.open_thread(request, thread)

        return {"is_closed": thread.is_closed}
    else:
        if value:
            raise PermissionDenied(
                pgettext(
                    "threads api", "You don't have permission to close this thread."
                )
            )
        else:
            raise PermissionDenied(
                pgettext(
                    "threads api", "You don't have permission to open this thread."
                )
            )


thread_patch_dispatcher.replace("is-closed", patch_is_closed)


def patch_is_hidden(request, thread, value):
    if value:
        allow_hide_thread(request.user_acl, thread)
        moderation.hide_thread(request, thread)
    else:
        allow_unhide_thread(request.user_acl, thread)
        moderation.unhide_thread(request, thread)

    return {"is_hidden": thread.is_hidden}


thread_patch_dispatcher.replace("is-hidden", patch_is_hidden)


def patch_best_answer(request, thread, value):
    try:
        post_id = int(value)
    except (TypeError, ValueError):
        raise PermissionDenied(pgettext("threads api", "A valid integer is required."))

    allow_mark_best_answer(request.user_acl, thread)

    post = get_object_or_404(thread.post_set, id=post_id)
    post.category = thread.category
    post.thread = thread

    allow_see_post(request.user_acl, post)
    allow_mark_as_best_answer(request.user_acl, post)

    if post.is_best_answer:
        raise PermissionDenied(
            pgettext(
                "threads api", "This post is already marked as thread's best answer."
            )
        )

    if thread.has_best_answer:
        allow_change_best_answer(request.user_acl, thread)

    thread.set_best_answer(request.user, post)
    thread.save()

    return {
        "best_answer": thread.best_answer_id,
        "best_answer_is_protected": thread.best_answer_is_protected,
        "best_answer_marked_on": thread.best_answer_marked_on,
        "best_answer_marked_by": thread.best_answer_marked_by_id,
        "best_answer_marked_by_name": thread.best_answer_marked_by_name,
        "best_answer_marked_by_slug": thread.best_answer_marked_by_slug,
    }


thread_patch_dispatcher.replace("best-answer", patch_best_answer)


def patch_unmark_best_answer(request, thread, value):
    try:
        post_id = int(value)
    except (TypeError, ValueError):
        raise PermissionDenied(pgettext("threads api", "A valid integer is required."))

    post = get_object_or_404(thread.post_set, id=post_id)
    post.category = thread.category
    post.thread = thread

    if not post.is_best_answer:
        raise PermissionDenied(
            pgettext(
                "threads api",
                "This post can't be unmarked because it's not currently marked as best answer.",
            )
        )

    allow_unmark_best_answer(request.user_acl, thread)
    thread.clear_best_answer()
    thread.save()

    return {
        "best_answer": None,
        "best_answer_is_protected": False,
        "best_answer_marked_on": None,
        "best_answer_marked_by": None,
        "best_answer_marked_by_name": None,
        "best_answer_marked_by_slug": None,
    }


thread_patch_dispatcher.remove("best-answer", patch_unmark_best_answer)


def patch_add_participant(request, thread, value: str):
    allow_add_participants(request.user_acl, thread)

    try:
        user_slug = slugify_username(value)
        if not user_slug:
            raise PermissionDenied(
                pgettext("threads api", "You have to enter new participant's username.")
            )
        participant = User.objects.get(slug=user_slug)
    except User.DoesNotExist:
        raise PermissionDenied(
            pgettext("threads api", "No user with this name exists.")
        )

    if participant in [p.user for p in thread.participants_list]:
        raise PermissionDenied(
            pgettext("threads api", "This user is already thread participant.")
        )

    participant_acl = useracl.get_user_acl(participant, request.cache_versions)
    allow_add_participant(request.user_acl, participant, participant_acl)
    add_participant(request, thread, participant)

    make_participants_aware(request.user, thread)
    participants = ThreadParticipantSerializer(thread.participants_list, many=True)

    return {"participants": participants.data}


thread_patch_dispatcher.add("participants", patch_add_participant)


def patch_remove_participant(request, thread, value):
    # pylint: disable=undefined-loop-variable
    try:
        user_id = int(value)
    except (ValueError, TypeError):
        raise PermissionDenied(pgettext("threads api", "A valid integer is required."))

    for participant in thread.participants_list:
        if participant.user_id == user_id:
            break
    else:
        raise PermissionDenied(pgettext("threads api", "Participant doesn't exist."))

    allow_remove_participant(request.user_acl, thread, participant.user)
    remove_participant(request, thread, participant.user)

    if len(thread.participants_list) == 1:
        return {"deleted": True}

    make_participants_aware(request.user, thread)
    participants = ThreadParticipantSerializer(thread.participants_list, many=True)

    return {"deleted": False, "participants": participants.data}


thread_patch_dispatcher.remove("participants", patch_remove_participant)


def patch_replace_owner(request, thread, value):
    # pylint: disable=undefined-loop-variable
    try:
        user_id = int(value)
    except (ValueError, TypeError):
        raise PermissionDenied(pgettext("threads api", "A valid integer is required."))

    for participant in thread.participants_list:
        if participant.user_id == user_id:
            if participant.is_owner:
                raise PermissionDenied(
                    pgettext("threads api", "This user already is thread owner.")
                )
            else:
                break
    else:
        raise PermissionDenied(pgettext("threads api", "Participant doesn't exist."))

    allow_change_owner(request.user_acl, thread)
    change_owner(request, thread, participant.user)

    make_participants_aware(request.user, thread)
    participants = ThreadParticipantSerializer(thread.participants_list, many=True)
    return {"participants": participants.data}


thread_patch_dispatcher.replace("owner", patch_replace_owner)


def thread_patch_endpoint(request, thread):
    old_title = thread.title
    old_is_hidden = thread.is_hidden
    old_is_unapproved = thread.is_unapproved
    old_category = thread.category

    response = thread_patch_dispatcher.dispatch(request, thread)

    # diff thread's state against pre-patch and resync category if necessary
    hidden_changed = old_is_hidden != thread.is_hidden
    unapproved_changed = old_is_unapproved != thread.is_unapproved
    category_changed = old_category != thread.category

    title_changed = old_title != thread.title
    if thread.category.last_thread_id != thread.pk:
        title_changed = False  # don't trigger resync on simple title change

    if hidden_changed or unapproved_changed or category_changed:
        thread.category.synchronize()
        thread.category.save()

        if category_changed:
            old_category.synchronize()
            old_category.save()
    elif title_changed:
        thread.category.last_thread_title = thread.title
        thread.category.last_thread_slug = thread.slug
        thread.category.save(update_fields=["last_thread_title", "last_thread_slug"])

    return response


def bulk_patch_endpoint(
    request, viewmodel
):  # pylint: disable=too-many-branches, too-many-locals
    serializer = BulkPatchSerializer(
        data=request.data, context={"settings": request.settings}
    )
    if not serializer.is_valid():
        return Response(serializer.errors, status=400)

    threads = clean_threads_for_patch(request, viewmodel, serializer.data["ids"])

    old_titles = [t.title for t in threads]
    old_is_hidden = [t.is_hidden for t in threads]
    old_is_unapproved = [t.is_unapproved for t in threads]
    old_category = [t.category_id for t in threads]

    response = thread_patch_dispatcher.dispatch_bulk(request, threads)

    new_titles = [t.title for t in threads]
    new_is_hidden = [t.is_hidden for t in threads]
    new_is_unapproved = [t.is_unapproved for t in threads]
    new_category = [t.category_id for t in threads]

    # sync titles
    if new_titles != old_titles:
        for i, t in enumerate(threads):
            if t.title != old_titles[i] and t.category.last_thread_id == t.pk:
                t.category.last_thread_title = t.title
                t.category.last_thread_slug = t.slug
                t.category.save(update_fields=["last_thread_title", "last_thread_slug"])

    # sync categories
    sync_categories = []

    if new_is_hidden != old_is_hidden:
        for i, t in enumerate(threads):
            if t.is_hidden != old_is_hidden[i] and t.category_id not in sync_categories:
                sync_categories.append(t.category_id)

    if new_is_unapproved != old_is_unapproved:
        for i, t in enumerate(threads):
            if (
                t.is_unapproved != old_is_unapproved[i]
                and t.category_id not in sync_categories
            ):
                sync_categories.append(t.category_id)

    if new_category != old_category:
        for i, t in enumerate(threads):
            if t.category_id != old_category[i]:
                if t.category_id not in sync_categories:
                    sync_categories.append(t.category_id)
                if old_category[i] not in sync_categories:
                    sync_categories.append(old_category[i])

    if sync_categories:
        for category in Category.objects.filter(id__in=sync_categories):
            category.synchronize()
            category.save()

    return response


def clean_threads_for_patch(request, viewmodel, threads_ids):
    threads = []
    for thread_id in sorted(set(threads_ids), reverse=True):
        try:
            threads.append(viewmodel(request, thread_id).unwrap())
        except (Http404, PermissionDenied):
            raise PermissionDenied(
                pgettext(
                    "threads api", "One or more threads to update could not be found."
                )
            )
    return threads


class BulkPatchSerializer(serializers.Serializer):
    ids = serializers.ListField(
        child=serializers.IntegerField(min_value=1), min_length=1
    )
    ops = serializers.ListField(
        child=serializers.DictField(), min_length=1, max_length=10
    )

    def validate_ids(self, data):
        limit = self.context["settings"].threads_per_page
        if len(data) > limit:
            message = npgettext(
                "threads api",
                "No more than %(limit)s thread can be updated at a single time.",
                "No more than %(limit)s threads can be updated at a single time.",
                limit,
            )
            raise ValidationError(message % {"limit": limit})
        return data



from django.db import migrations

from .. import SOCIALAUTH_CACHE
from ...cache.operations import StartCacheVersioning


class Migration(migrations.Migration):
    dependencies = [
        ("misago_socialauth", "0001_initial"),
        ("misago_cache", "0001_initial"),
    ]

    operations = [StartCacheVersioning(SOCIALAUTH_CACHE)]



from django.apps import AppConfig


class MisagoThreadsConfig(AppConfig):
    name = "misago.threads"
    label = "misago_threads"
    verbose_name = "Misago Threads"

    def ready(self):
        from . import signals as _



from functools import reduce
from typing import Any, Generic, List, TypeVar, cast

Action = TypeVar("Action")
Filter = TypeVar("Filter")


class ActionHook(Generic[Action]):
    __slots__ = ("_actions_first", "_actions_last", "_cache")

    _actions_first: List[Action]
    _actions_last: List[Action]
    _cache: List[Action] | None

    def __init__(self):
        self._actions_first = []
        self._actions_last = []
        self._cache = None

    def __bool__(self) -> bool:
        return bool(self._actions_first or self._actions_last)

    def append_action(self, action: Action):
        self._actions_last.append(action)
        self.invalidate_cache()

    def prepend_action(self, action: Action):
        self._actions_first.insert(0, action)
        self.invalidate_cache()

    def invalidate_cache(self):
        self._cache = None

    def __call__(self, *args, **kwargs) -> List[Any]:
        if self._cache is None:
            self._cache = self._actions_first + self._actions_last
        if not self._cache:
            return []

        return [action(*args, **kwargs) for action in self._cache]


class FilterHook(Generic[Action, Filter]):
    __slots__ = ("_filters_first", "_filters_last", "_cache")

    _filters_first: List[Filter]
    _filters_last: List[Filter]
    _cache: Action | None

    def __init__(self):
        self._filters_first = []
        self._filters_last = []
        self._cache = None

    def __bool__(self) -> bool:
        return bool(self._filters_first or self._filters_last)

    def append_filter(self, filter_: Filter):
        self._filters_last.append(filter_)
        self.invalidate_cache()

    def prepend_filter(self, filter_: Filter):
        self._filters_first.insert(0, filter_)
        self.invalidate_cache()

    def invalidate_cache(self):
        self._cache = None

    def get_reduced_action(self, action: Action) -> Action:
        def reduce_filter(action: Action, next_filter: Filter) -> Action:
            def reduced_filter(*args, **kwargs):
                return next_filter(action, *args, **kwargs)

            return cast(Action, reduced_filter)

        filters = self._filters_first + self._filters_last
        return reduce(reduce_filter, filters, action)

    def __call__(self, action: Action, *args, **kwargs):
        if self._cache is None:
            self._cache = self.get_reduced_action(action)

        return self._cache(*args, **kwargs)  # type: ignore



from django.db import models
from django.utils import timezone

from ..conf import settings


class Subject(models.Model):
    sub = models.CharField(max_length=36, primary_key=True)  # UUID4 str length
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    created_on = models.DateTimeField(default=timezone.now)
    last_used_on = models.DateTimeField(default=timezone.now)



import pytest
from django.urls import reverse

from ...test import GraphQLTestClient


@pytest.fixture
def admin_graphql_client(admin_client):
    return GraphQLTestClient(admin_client, reverse("misago:admin:graphql:index"))



from ...acl.test import patch_user_acl
from ..test import AuthenticatedUserTestCase


class UserSignatureTests(AuthenticatedUserTestCase):
    """tests for user signature RPC (POST to /api/users/1/signature/)"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/signature/" % self.user.pk

    @patch_user_acl({"can_have_signature": 0})
    def test_signature_no_permission(self):
        """edit signature api with no ACL returns 403"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You don't have permission to change signature."},
        )

    @patch_user_acl({"can_have_signature": 1})
    def test_signature_locked(self):
        """locked edit signature returns 403"""
        self.user.is_signature_locked = True
        self.user.signature_lock_user_message = "Your siggy is banned."
        self.user.save()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": "Your signature is locked. You can't change it.",
                "reason": "<p>Your siggy is banned.</p>",
            },
        )

    @patch_user_acl({"can_have_signature": 1})
    def test_get_signature(self):
        """GET to api returns json with no signature"""
        self.user.is_signature_locked = False
        self.user.save()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        self.assertFalse(response.json()["signature"])

    @patch_user_acl({"can_have_signature": 1})
    def test_post_empty_signature(self):
        """empty POST empties user signature"""
        self.user.is_signature_locked = False
        self.user.save()

        response = self.client.post(self.link, data={"signature": ""})
        self.assertEqual(response.status_code, 200)

        self.assertFalse(response.json()["signature"])

    @patch_user_acl({"can_have_signature": 1})
    def test_post_too_long_signature(self):
        """too long new signature errors"""
        self.user.is_signature_locked = False
        self.user.save()

        response = self.client.post(self.link, data={"signature": "abcd" * 1000})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Signature is too long."})

    @patch_user_acl({"can_have_signature": 1})
    def test_post_good_signature(self):
        """POST with good signature changes user signature"""
        self.user.is_signature_locked = False
        self.user.save()

        response = self.client.post(self.link, data={"signature": "Hello, **bros**!"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json()["signature"]["html"], "<p>Hello, <strong>bros</strong>!</p>"
        )
        self.assertEqual(response.json()["signature"]["plain"], "Hello, **bros**!")

        self.reload_user()
        self.assertEqual(
            self.user.signature_parsed, "<p>Hello, <strong>bros</strong>!</p>"
        )



from contextlib import ContextDecorator, ExitStack, contextmanager
from unittest.mock import patch

from .useracl import get_user_acl

__all__ = ["patch_user_acl"]


class patch_user_acl(ContextDecorator, ExitStack):
    """Testing utility that patches get_user_acl results

    Can be used as decorator or context manager.

    Patch should be a dict or callable.
    """

    _acl_patches = []

    def __init__(self, acl_patch):
        super().__init__()
        self.acl_patch = acl_patch

    def patched_get_user_acl(self, user, cache_versions):
        user_acl = get_user_acl(user, cache_versions)
        self.apply_acl_patches(user, user_acl)
        return user_acl

    def apply_acl_patches(self, user, user_acl):
        for acl_patch in self._acl_patches:
            self.apply_acl_patch(user, user_acl, acl_patch)

    def apply_acl_patch(self, user, user_acl, acl_patch):
        if callable(acl_patch):
            acl_patch(user, user_acl)
        else:
            user_acl.update(acl_patch)

    def __enter__(self):
        super().__enter__()
        self.enter_context(self.enable_acl_patch())
        self.enter_context(self.patch_user_acl())

    @contextmanager
    def enable_acl_patch(self):
        try:
            self._acl_patches.append(self.acl_patch)
            yield
        finally:
            self._acl_patches.pop(-1)

    def patch_user_acl(self):
        return patch(
            "misago.acl.useracl.get_user_acl", side_effect=self.patched_get_user_acl
        )



import os

import pytest
from django.core.files.uploadedfile import UploadedFile
from django.urls import reverse

from ....test import assert_has_error_message

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def css_file():
    return os.path.join(TESTS_DIR, "css", "test.css")


@pytest.fixture
def other_file():
    return os.path.join(TESTS_DIR, "images", "test.png")


@pytest.fixture
def hashed_css_file():
    return os.path.join(TESTS_DIR, "css", "test.4846cb3b.css")


@pytest.fixture
def upload(admin_client):
    def post_upload(theme, asset_files=None):
        url = reverse("misago:admin:themes:upload-css", kwargs={"pk": theme.pk})
        if asset_files is not None:
            data = asset_files if isinstance(asset_files, list) else [asset_files]
        else:
            data = ""
        return admin_client.post(url, {"assets": data})

    return post_upload


def test_css_file_can_be_uploaded(upload, theme, css_file):
    with open(css_file) as fp:
        upload(theme, fp)
        assert theme.css.exists()


def test_multiple_css_files_can_be_uploaded_at_once(
    upload, theme, css_file, hashed_css_file
):
    with open(css_file) as fp1:
        with open(hashed_css_file) as fp2:
            upload(theme, [fp1, fp2])
            assert theme.css.exists()
            assert theme.css.count() == 2


def test_css_files_uploaded_one_after_another_are_ordered(
    upload, theme, css_file, hashed_css_file
):
    with open(css_file) as fp:
        upload(theme, fp)
    first_css = theme.css.last()
    assert first_css.name == str(css_file).split("/")[-1]
    assert first_css.order == 0

    with open(hashed_css_file) as fp:
        upload(theme, fp)
    last_css = theme.css.last()
    assert last_css.name == str(hashed_css_file).split("/")[-1]
    assert last_css.order == 1


def test_multiple_css_files_uploaded_at_once_are_ordered(
    upload, theme, css_file, hashed_css_file
):
    with open(css_file) as fp1:
        with open(hashed_css_file) as fp2:
            upload(theme, [fp1, fp2])

    assert list(theme.css.values_list("name", flat=True)) == [
        str(css_file).split("/")[-1],
        str(hashed_css_file).split("/")[-1],
    ]
    assert list(theme.css.values_list("order", flat=True)) == [0, 1]


def test_uploaded_file_is_rejected_if_its_not_css_file(upload, theme, other_file):
    with open(other_file, "rb") as fp:
        upload(theme, fp)
        assert not theme.css.exists()


def test_error_message_is_set_if_uploaded_file_is_not_css(upload, theme, other_file):
    with open(other_file, "rb") as fp:
        response = upload(theme, fp)
        assert_has_error_message(response)


def test_if_some_of_uploaded_files_are_incorrect_only_css_files_are_added_to_theme(
    upload, theme, css_file, other_file
):
    with open(css_file) as fp1:
        with open(other_file, "rb") as fp2:
            upload(theme, [fp1, fp2])
            assert theme.css.exists()
            assert theme.css.count() == 1

    css = theme.css.last()
    expected_filename = str(css_file).split("/")[-1]
    assert css.name == expected_filename


def test_css_file_is_uploaded_to_theme_directory(upload, theme, css_file):
    with open(css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    assert theme.dirname in str(css.source_file)


def test_css_file_name_is_set_as_asset_name(upload, theme, css_file):
    with open(css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    expected_filename = str(css_file).split("/")[-1]
    assert css.name == expected_filename


def test_hash_is_added_to_uploaded_css_file_name(
    upload, theme, css_file, hashed_css_file
):
    with open(css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    filename = str(css.source_file.path).split("/")[-1]
    expected_filename = str(hashed_css_file).split("/")[-1]
    assert filename == expected_filename


def test_hash_is_set_on_css_source_asset(upload, theme, css_file):
    with open(css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    assert css.source_hash


def test_css_file_name_is_preserved_if_it_already_contains_correct_hash(
    upload, theme, hashed_css_file
):
    with open(hashed_css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    filename = str(css.source_file.path).split("/")[-1]
    expected_filename = str(hashed_css_file).split("/")[-1]
    assert filename == expected_filename


def test_new_hash_is_added_to_css_file_name_if_it_contains_incorrect_hash(
    upload, theme
):
    incorrectly_hashed_css_file = os.path.join(TESTS_DIR, "css", "test.0046cb3b.css")
    with open(incorrectly_hashed_css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    filename = str(css.source_file.path).split("/")[-1]
    assert css.source_hash in filename


def test_newly_uploaded_css_file_replaces_old_one_if_file_names_are_same(
    upload, theme, css_file
):
    with open(css_file) as fp:
        upload(theme, fp)
    original_css = theme.css.get()

    with open(os.path.join(TESTS_DIR, "css", "test-changed.css")) as fp:
        size = len(fp.read())
        fp.seek(0)
        upload(
            theme, UploadedFile(fp, name="test.css", content_type="text/css", size=size)
        )
    updated_css = theme.css.last()

    assert updated_css.name == original_css.name
    assert updated_css.source_hash != original_css.source_hash
    assert theme.css.count() == 1


def test_newly_uploaded_css_file_reuses_replaced_file_order_if_names_are_same(
    upload, theme, css_file, hashed_css_file
):
    with open(css_file) as fp:
        upload(theme, fp)
    original_css = theme.css.last()

    with open(hashed_css_file) as fp:
        upload(theme, fp)

    with open(os.path.join(TESTS_DIR, "css", "test-changed.css")) as fp:
        size = len(fp.read())
        fp.seek(0)
        upload(
            theme, UploadedFile(fp, name="test.css", content_type="text/css", size=size)
        )

    updated_css = theme.css.get(order=original_css.order)
    assert updated_css.name == original_css.name


def test_if_uploaded_css_file_contains_no_image_urls_rebuild_flag_is_not_set(
    upload, theme, css_file
):
    with open(css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    assert not css.source_needs_building


def test_if_uploaded_css_file_contains_image_url_it_has_rebuild_flag_set(upload, theme):
    css_file = os.path.join(TESTS_DIR, "css", "test.needs-build.css")
    with open(css_file) as fp:
        upload(theme, fp)

    css = theme.css.last()
    assert css.source_needs_building


def test_uploading_css_file_triggers_css_build(
    upload, theme, css_file, mock_build_theme_css
):
    with open(css_file) as fp:
        upload(theme, fp)

    mock_build_theme_css.assert_called_once_with(theme.pk)


def test_error_message_is_set_if_no_css_file_was_uploaded(upload, theme):
    response = upload(theme)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_upload_css_file_to_default_theme(
    upload, default_theme
):
    response = upload(default_theme)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_upload_css_file_to_nonexisting_theme(
    upload, nonexisting_theme
):
    response = upload(nonexisting_theme)
    assert_has_error_message(response)
